#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Module handling the start up of the fitting procedure."""
import sys
import argparse
import os
import time
import importlib

import numpy as np
from tkinter import filedialog
from tkinter import Tk
from inputimeout import inputimeout, TimeoutOccurred

import matplotlib.pyplot as plt

import h5py
import pyopencl as cl
import pyopencl.array as clarray


from pyqmri._helper_fun import _goldcomp as goldcomp
from pyqmri._helper_fun._est_coils import est_coils
from pyqmri._helper_fun import _utils as utils
from pyqmri.solver import CGSolver
from pyqmri.irgn import IRGNOptimizer
np.seterr(divide='ignore', invalid='ignore')

np.seterr(divide='ignore')


def _choosePlatform(myargs, par):
    platforms = cl.get_platforms()
    use_GPU = False
    par["Platform_Indx"] = None
    dev_ind = []
    inp = None
    if myargs.use_GPU is True:
        for j, platform in enumerate(platforms):
            if platform.get_devices(device_type=cl.device_type.GPU):
                print("GPU OpenCL platform <%s> found "
                      "with %i device(s) and OpenCL-version <%s>"
                      % (str(platform.get_info(cl.platform_info.NAME)),
                         len(platform.get_devices(
                             device_type=cl.device_type.GPU)),
                         str(platform.get_info(cl.platform_info.VERSION))))
                use_GPU = True
                dev_ind.append(j)
        if len(dev_ind) > 1:
            while True:
                try:
                    inp = inputimeout(
                        (str(len(dev_ind)) +
                         " GPU platforms found."+
                         " Defaulting to (0) in 4 seconds.\n"), 
                        4)
                except:
                    inp = None
                if inp is None:
                    break
                else:
                    try:
                        inp = int(inp)
                    except:
                        print("Please specify an integer.")
                        continue
                    if inp >= len(dev_ind) or inp < 0:
                        print("Integer needs to be in range 0-"
                              +str(len(dev_ind)-1))
                        continue
                break
        if inp:
            par["Platform_Indx"] = dev_ind[inp]
        else:
            par["Platform_Indx"] = dev_ind[0]
     
    if not use_GPU:
        if myargs.use_GPU:
            print("No GPU OpenCL platform found. Falling back to CPU.")
        for j, platform in enumerate(platforms):
            if platform.get_devices(device_type=cl.device_type.CPU):
                print("CPU OpenCL platform <%s> found "
                      "with %i device(s) and OpenCL-version <%s>"
                      % (str(platform.get_info(cl.platform_info.NAME)),
                         len(platform.get_devices(
                             device_type=cl.device_type.CPU)),
                         str(platform.get_info(cl.platform_info.VERSION))))
                use_GPU = False
                par["Platform_Indx"] = j
        if par["Platform_Indx"] is None:
            raise(ValueError("No OpenCL device found."))   

    par["use_GPU"] = use_GPU
    return platforms


def _precoompFFT(data, par):

    trans_x = False
    trans_y = False
    trans_z = False

    if par["is3D"]:
        full_dimY = (np.all(np.abs(data[0, 0, 0, :, 0])) or
                     np.all(np.abs(data[0, 0, 0, :, 1])) or
                     np.all(np.abs(data[0, 0, 1, :, 0])) or
                     np.all(np.abs(data[0, 0, 1, :, 1])))

        full_dimX = (np.all(np.abs(data[0, 0, 0, 0, :])) or
                     np.all(np.abs(data[0, 0, 0, 1, :])) or
                     np.all(np.abs(data[0, 0, 1, 0, :])) or
                     np.all(np.abs(data[0, 0, 1, 1, :])))

        full_dimZ = (np.all(np.abs(data[0, 0, :, 0, 0:])) or
                     np.all(np.abs(data[0, 0, :, 1, 1])) or
                     np.all(np.abs(data[0, 0, :, 0, 1])) or
                     np.all(np.abs(data[0, 0, :, 1, 0])))
    else:
        full_dimY = (np.all(np.abs(data[0, 0, 0, :, 0])) or
                     np.all(np.abs(data[0, 0, 0, :, 1])))
        full_dimX = (np.all(np.abs(data[0, 0, 0, 0, :])) or
                     np.all(np.abs(data[0, 0, 0, 1, :])))
        full_dimZ = True

    if full_dimZ:
        if par["is3D"]:
            print("Image Dimensions Z seems fully sampled. "
                  "Precompute FFT along Z")
            data = np.fft.ifft(data, axis=-3, norm='ortho')
        trans_z = True

    if full_dimY:
        print("Image Dimensions Y seems fully sampled. "
              "Precompute FFT along Y")
        data = np.fft.ifft(data, axis=-2, norm='ortho')
        trans_y = True

    if full_dimX:
        print("Image Dimensions X seems fully sampled. "
              "Precompute FFT along X")
        data = np.fft.ifft(data, axis=-1, norm='ortho')
        trans_x = True

    if trans_x:
        if trans_y and trans_z:
            par["fft_dim"] = None
        elif not trans_y and not trans_z:
            data = np.require(
                data.transpose(0, 1, -1, -2, -3),
                requirements='C')
            par["C"] = np.require(
                par["C"].transpose(0, -1, -2, -3),
                requirements='C')
            par["mask"] = np.require(
                par["mask"].T,
                requirements='C')
            dimX = par["dimX"]
            NSlice = par["NSlice"]
            par["dimX"] = NSlice
            par["NSlice"] = dimX
            par["N"] = NSlice
            par["transpXY"] = True
            par["fft_dim"] = [-2, -1]
        elif not trans_z and trans_y:
            data = np.require(
                data.transpose(0, 1, -1, -2, -3),
                requirements='C')
            par["C"] = np.require(
                par["C"].transpose(0, -1, -2, -3),
                requirements='C')
            par["mask"] = np.require(
                par["mask"].T,
                requirements='C')
            dimX = par["dimX"]
            NSlice = par["NSlice"]
            par["dimX"] = NSlice
            par["NSlice"] = dimX
            par["N"] = NSlice
            par["transpXY"] = True
            par["fft_dim"] = [-1]
        else:
            data = np.require(
                np.moveaxis(data, -1, -2),
                requirements='C')
            par["C"] = np.require(
                np.moveaxis(par["C"], -1, -2),
                requirements='C')
            par["mask"] = np.require(
                np.moveaxis(par["mask"], -1, -2),
                requirements='C')
            dimX = par["dimX"]
            dimY = par["dimY"]
            par["dimX"] = dimY
            par["dimY"] = dimX
            par["N"] = dimY
            par["Nproj"] = dimX
            par["transpXY"] = True
            par["fft_dim"] = [-1]
    else:
        if trans_y and trans_z:
            par["fft_dim"] = [-1]
        elif not trans_y and not trans_z:
            par["fft_dim"] = [-3, -2, -1]
        elif not trans_z and trans_y:
            data = np.require(
                np.moveaxis(data, -2, -3),
                requirements='C')
            par["C"] = np.require(
                np.moveaxis(par["C"], -2, -3),
                requirements='C')
            par["mask"] = np.require(
                np.moveaxis(par["mask"], -2, -3),
                requirements='C')
            dimY = par["dimY"]
            NSlice = par["NSlice"]
            par["dimY"] = NSlice
            par["NSlice"] = dimY
            par["Nproj"] = NSlice
            par["transpXY"] = True
            par["fft_dim"] = [-2, -1]
        else:
            par["fft_dim"] = [-2, -1]

    return np.require(data.astype(par["DTYPE"]),
                      requirements='C')


def _setupOCL(myargs, par):
    platforms = _choosePlatform(myargs, par)
    par["ctx"] = []
    par["queue"] = []
    if par["use_GPU"]:
        queue_par = cl.command_queue_properties.OUT_OF_ORDER_EXEC_MODE_ENABLE
    else:
        queue_par = None
    if isinstance(myargs.devices, int):
        myargs.devices = [myargs.devices]
    if myargs.streamed:
        if len(myargs.devices) == 1 and myargs.devices[0] == -1:
            num_dev = []
            for j in range(len(platforms[par["Platform_Indx"]].get_devices())):
                num_dev.append(j)
            par["num_dev"] = num_dev
        else:
            num_dev = myargs.devices
            par["num_dev"] = num_dev
    else:
        num_dev = myargs.devices
        par["num_dev"] = num_dev
    for device in num_dev:
        dev = []
        dev.append(platforms[par["Platform_Indx"]].get_devices()[device])
        tmpxtx = cl.Context(dev)
        par["ctx"].append(tmpxtx)
        for j in range(4):
            par["queue"].append(
                cl.CommandQueue(
                   tmpxtx,
                   platforms[par["Platform_Indx"]].get_devices()[device],
                   properties=queue_par
                   )
                )


def _genImages(myargs, par, data, off):
    if not myargs.usecg:
        FFT = utils.NUFFT(par, trafo=myargs.trafo, SMS=myargs.sms)

        def nFTH(x, fft, par):
            siz = np.shape(x)
            result = np.zeros(
                (par["NScan"], par["NC"], par["NSlice"],
                 par["dimY"], par["dimX"]), dtype=par["DTYPE"])
            tmp_result = clarray.empty(fft.queue,
                                       (1, 1, par["NSlice"],
                                        par["dimY"], par["dimX"]),
                                       dtype=par["DTYPE"])
            start = time.time()
            for j in range(siz[0]):
                for k in range(siz[1]):
                    inp = clarray.to_device(fft.queue,
                                            np.require(x[j, k, ...]
                                                        [None, None, ...],
                                                       requirements='C'))
                    fft.FFTH(tmp_result, inp, scan_offset=j).wait()
                    result[j, k, ...] = np.squeeze(tmp_result.get())
            end = time.time()-start
            print("FT took %f s" % end)
            return result
        images = np.require(np.sum(nFTH(data, FFT, par) *
                                   (np.conj(par["C"])), axis=1),
                            requirements='C')
        del FFT, nFTH

    else:
        tol = 1e-16
        par_scans = 4
        lambd = 1e-1
        if "images" not in list(par["file"].keys()):
            images = np.zeros((par["NScan"],
                               par["NSlice"],
                               par["dimY"],
                               par["dimX"]), dtype=par["DTYPE"])
            if par["NScan"]/par_scans >= 1:
                cgs = CGSolver(par, par_scans, myargs.trafo, myargs.sms)
                for j in range(int(par["NScan"]/par_scans)):
                    if par["NSlice"] == 1:
                        images[par_scans*j:par_scans*(j+1), ...] = cgs.run(
                            data[par_scans*j:par_scans*(j+1), ...],
                            tol=tol, lambd=lambd,
                            scan_offset=par_scans*j)[:, None, ...]
                    else:
                        images[par_scans*j:par_scans*(j+1), ...] = cgs.run(
                            data[par_scans*j:par_scans*(j+1), ...],
                            tol=tol, lambd=lambd, scan_offset=par_scans*j)
                del cgs
            if np.mod(par["NScan"], par_scans):
                cgs = CGSolver(par, np.mod(par["NScan"], par_scans),
                               myargs.trafo, myargs.sms)
                if par["NSlice"] == 1:
                    if np.mod(par["NScan"], par_scans) == 1:
                        images[-np.mod(par["NScan"], par_scans):, ...] = \
                            cgs.run(
                                data[-np.mod(par["NScan"], par_scans):, ...],
                                tol=tol, lambd=lambd,
                                scan_offset=par["NScan"]-np.mod(
                                    par["NScan"], par_scans))
                    else:
                        images[-np.mod(par["NScan"], par_scans):, ...] = \
                            cgs.run(
                                data[-np.mod(par["NScan"], par_scans):, ...],
                                tol=tol, lambd=lambd,
                                scan_offset=par["NScan"]-np.mod(
                                    par["NScan"], par_scans))[:, None, ...]
                else:
                    images[-np.mod(par["NScan"], par_scans):, ...] = \
                        cgs.run(
                            data[-np.mod(par["NScan"], par_scans):, ...],
                            tol=tol, lambd=lambd,
                            scan_offset=par["NScan"]-np.mod(
                                    par["NScan"], par_scans))
                del cgs
            par["file"].create_dataset("images", images.shape,
                                       dtype=par["DTYPE"], data=images)
        else:
            images = par["file"]['images']
            if images.shape[1] < par["NSlice"]:
                del par["file"]["images"]
                images = _genImages(myargs, par, data, off)
            else:
                print("Using precomputed images")
                slices_images = par["file"]['images'][()].shape[1]
                images = \
                    par["file"]['images'][
                      :,
                      int(slices_images / 2) - int(
                        np.floor((par["NSlice"]) / 2)) + off:int(
                          slices_images / 2) + int(
                            np.ceil(par["NSlice"] / 2)) + off,
                      ...].astype(par["DTYPE"])
    return images


def _estScaleNorm(myargs, par, images, data):

    # if myargs.trafo:
    #     center = int(par["N"]*0.1)
    #     ind = np.zeros((par["N"]), dtype=bool)
    #     ind[int(par["N"]/2-center):int(par["N"]/2+center)] = 1
    #     inds = np.fft.fftshift(ind)
    #     dims = tuple(range(data.ndim - 3)) + (-2,)
    #     sig = np.max(
    #         np.sum(data[..., ind], dims) *
    #         np.conj(np.sum(data[..., ind], dims)))
    #     noise = np.std(
    #         np.sum(data[..., inds], dims) *
    #         np.conj(np.sum(data[..., inds], dims)))

    # else:
    #     centerX = int(par["dimX"]*0.1)
    #     centerY = int(par["dimY"]*0.1)
    #     if myargs.is3Ddata:
    #         centerZ = int(par["NSlice"]*0.1)
    #     else:
    #         centerZ = int(par["NSlice"]/2)

    #     ind = np.zeros((par["NSlice"], par["dimY"], par["dimX"]), dtype=bool)
    #     ind[int(par["NSlice"]/2-centerZ):int(par["NSlice"]/2+centerZ),
    #         int(par["dimY"]/2-centerY):int(par["dimY"]/2+centerY),
    #         int(par["dimX"]/2-centerX):int(par["dimX"]/2+centerX)] = 1
    #     if par["fft_dim"] is not None:
    #         for shiftdim in par["fft_dim"]:
    #             ind = np.fft.fftshift(ind, axes=shiftdim)
    #         sig = np.sum(
    #             np.abs(data[..., ind])**2)
    #         noise = np.sum(
    #             np.abs(data[..., ~ind])**2)
    #     else:
    #         tmp = np.fft.fft2(data, norm='ortho')
    #         ind = np.fft.fftshift(ind)
    #         sig = np.sum(
    #             np.abs(tmp[..., ind])**2)
    #         noise = np.sum(
    #             np.abs(tmp[..., ~ind])**2)

    # SNR_est = (np.abs(sig/noise))
    # par["SNR_est"] = SNR_est
    # print("Estimated SNR from kspace", SNR_est)
    dscale = par["DTYPE_real"](1 / np.linalg.norm(np.abs(data)))
    print("Data scale: ", dscale)
    par["dscale"] = dscale
    images = images*dscale
    data = data*dscale

    return data, images


def _readInput(myargs, par):
    if myargs.file == '':
        select_file = True
        while select_file is True:
            root = Tk()
            root.withdraw()
            root.update()
            file = filedialog.askopenfilename()
            root.destroy()
            if file and \
               not file.endswith((('.h5'), ('.hdf5'))):
                print("Please specify a h5 file. Press cancel to exit.")
            elif not file:
                print("Exiting...")
                sys.exit()
            else:
                select_file = False
    else:
        if not myargs.file.endswith((('.h5'), ('.hdf5'))):
            print("Please specify a h5 file. ")
            sys.exit()
        file = myargs.file
    name = os.path.normpath(file)
    par["fname"] = name.split(os.sep)[-1].split(".")[0]

    if myargs.outdir == '':
        outdir = os.sep.join(name.split(os.sep)[:-1]) + os.sep + \
            "PyQMRI_out" + \
            os.sep + myargs.sig_model + os.sep + \
            time.strftime("%Y-%m-%d  %H-%M-%S") + os.sep
        if not os.sep.join(name.split(os.sep)[:-1]):
            outdir = '.'+outdir
    else:
        outdir = myargs.outdir + os.sep + "PyQMRI_out" + \
            os.sep + myargs.sig_model + os.sep + par["fname"] + os.sep + \
            time.strftime("%Y-%m-%d  %H-%M-%S") + os.sep
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    par["outdir"] = outdir
    par["file"] = h5py.File(file, 'a')


def _read_data_from_file(par, myargs):
    reco_Slices = myargs.slices
    dimX, dimY, NSlice = ((par["file"].attrs['image_dimensions']).astype(int))
    if reco_Slices == -1:
        reco_Slices = NSlice
    off = 0

    if myargs.sms or myargs.is3Ddata:
        data = par["file"]['real_dat'][()].astype(par["DTYPE"])\
               + 1j*par["file"]['imag_dat'][()].astype(par["DTYPE"])
    else:
        data = par["file"]['real_dat'][
          ...,
          int(NSlice/2)-int(np.floor((reco_Slices)/2))+off:
          int(NSlice/2)+int(np.ceil(reco_Slices/2))+off, :, :
              ].astype(par["DTYPE"])\
               + 1j*par["file"]['imag_dat'][
          ...,
          int(NSlice/2)-int(np.floor((reco_Slices)/2))+off:
          int(NSlice/2)+int(np.ceil(reco_Slices/2))+off, :, :
              ].astype(par["DTYPE"])

    dimreduction = 0
    if myargs.trafo:
        if "real_traj" in par["file"].keys():
            par["traj"] = np.stack((
                par["file"]['imag_traj'][()].astype(par["DTYPE_real"]),
                par["file"]['real_traj'][()].astype(par["DTYPE_real"])),
                axis=-1)
        else:
            par["traj"] = par["file"]['traj'][()].astype(par["DTYPE_real"])

        if "dcf" in par["file"].keys():
            par["dcf"] = np.sqrt(
                par["file"]["dcf"][()].astype(par["DTYPE_real"]))
        else:
            par["dcf"] = np.sqrt(goldcomp.cmp(
                             par["traj"]))
        par["dcf"] = np.require(np.abs(par["dcf"]),
                                par["DTYPE_real"], requirements='C')
                    
        # Check if traj is scaled
        max_traj_val = np.max(np.abs(par["traj"]))
        if  np.allclose(max_traj_val, 0.5, rtol=1e-1):
           par["traj"] *= dimX
        elif np.allclose(max_traj_val, 1, rtol=1e-1):
           par["traj"] *= dimX/2
        
        overgrid_factor_a = (1/np.linalg.norm(
            par["traj"][..., -2, :]-par["traj"][..., -1, :], axis=-1))
        overgrid_factor_b = (1/np.linalg.norm(
            par["traj"][..., 0, :]-par["traj"][..., 1, :], axis=-1))
        par["ogf"] = np.min((overgrid_factor_a,
                             overgrid_factor_b))
        print("Estimated OGF: ", par["ogf"])
        par["traj"] *= par["ogf"]
    else:
        par["traj"] = None
        par["dcf"] = None

    #### Need to rework this as it might introduce unknown errors in
    #### Cartesian data!
    if np.max(utils.prime_factors(data.shape[-1])) > 13:
        print("Samples along the spoke need to have their largest prime factor"
              " to be 13 or lower. Finding next smaller grid.")
        dimreduction = 2
        while np.max(utils.prime_factors(data.shape[-1]-dimreduction)) > 13:
            dimreduction += 2
        print('Decrease grid size by %i' % dimreduction)
        dimX -= dimreduction
        dimY -= dimreduction
        if myargs.trafo:
            data = np.require(data[..., int(dimreduction/2):
                                   data.shape[-1]-int(dimreduction/2)],
                              requirements='C')
            par["traj"] = np.require(par["traj"][
                ..., int(dimreduction/2):
                par["traj"].shape[-1]-int(dimreduction/2)], requirements='C')
            par["dcf"] = np.sqrt(goldcomp.cmp(par["traj"]))
            par["dcf"] = np.require(np.abs(par["dcf"]),
                                    par["DTYPE_real"], requirements='C')
        else:
            data = np.require(data[...,
                                   int(dimreduction/2):
                                   data.shape[-1]-int(dimreduction/2),
                                   int(dimreduction/2):
                                   data.shape[-1]-int(dimreduction/2)],
                              requirements='C')
    return data, dimX, dimY, NSlice, reco_Slices, dimreduction, off


def _read_flip_angle_correction_data(par, myargs, dimreduction, reco_Slices):
    if "fa_corr" in list(par["file"].keys()):
        NSlice_fa, _, _ = par["file"]['fa_corr'][()].shape
    elif "interpol_fa" in list(par["file"].keys()):
        NSlice_fa, _, _ = par["file"]['interpol_fa'][()].shape

    par["fa_corr"] = np.flip(
        par["file"]['fa_corr'][()].astype(par["DTYPE"]),
        0)[
          int(NSlice_fa/2)-int(np.floor((reco_Slices)/2)):
          int(NSlice_fa/2)+int(np.ceil(reco_Slices/2)),
          ...]
    par["fa_corr"][par["fa_corr"] == 0] = 1
    par["fa_corr"] = par["fa_corr"][
       ...,
       int(dimreduction/2):par["fa_corr"].shape[-2]-int(dimreduction/2),
       int(dimreduction/2):par["fa_corr"].shape[-1]-int(dimreduction/2)]
    par["fa_corr"] = np.require((np.transpose(par["fa_corr"], (0, 2, 1))),
                                requirements='C')


def _check_data_size(par, data, sigmodel):
    if data.ndim == 5:
        pass
    elif data.ndim == 4 and "IRLL" in sigmodel:
        Nproj_new = 13
        print("4D Data passed and IRLL model used. Reordering Projections "
              "into "+str(Nproj_new)+" Spokes/Frame")
        [NC, reco_Slices, Nproj, N] = data.shape
        NScan = np.floor_divide(Nproj, Nproj_new)
        par["Nproj_measured"] = Nproj
        Nproj = Nproj_new
        data = np.require(np.transpose(np.reshape(data[..., :Nproj*NScan, :],
                                                  (NC, reco_Slices, NScan,
                                                   Nproj, N)),
                                       (2, 0, 1, 3, 4)), requirements='C')
        par["traj"] = np.require(np.reshape(par["traj"][:Nproj*NScan, :],
                                            (NScan, Nproj, N)),
                                 requirements='C')
        par["dcf"] = np.sqrt(goldcomp.cmp(par["traj"]))
        par["dcf"] = np.require(np.abs(par["dcf"]), par["DTYPE_real"],
                                requirements='C')
    elif data.ndim == 4 and "ImageReco" in sigmodel:
        data = data[None]
    else:
        raise ValueError(
            "Wrong data dimension / model incompatible.")
    return data


def _populate_par_w_sequencepar(par,
                                datashape,
                                imagedims,
                                SMS,
                                imagespace,
                                streamded,
                                par_slices,
                                trafo):
    for att in par["file"].attrs:
        par[att] = par["file"].attrs[att]
    par["NC"] = datashape[1]
    par["dimY"] = imagedims[-2]
    par["dimX"] = imagedims[-1]
    if SMS:
        par["NSlice"] = imagedims[0]
        par["packs"] = int(par["packs"])
        par["numofpacks"] = int(
            imagedims[0]/(int(par["packs"])*int(par["MB"])))
    else:
        if par["is3D"]:
            par["NSlice"] = imagedims[0]
        else:
            par["NSlice"] = datashape[2]
        par["packs"] = 1
        par["MB"] = 1
    par["NScan"] = datashape[0]
    par["N"] = datashape[-1]
    par["Nproj"] = datashape[-2]
    par["imagespace"] = imagespace
    if par["is3D"]:
        par["fft_dim"] = (-3, -2, -1)
    else:
        par["fft_dim"] = (-2, -1)
    if streamded:
        par["par_slices"] = par_slices
        par["overlap"] = 1
    else:
        par["par_slices"] = par["NSlice"]
        par["overlap"] = 0
    par["transpXY"] = False


def _import_sigmodel(modelpath):
    sig_model_path = os.path.normpath(modelpath)
    if len(sig_model_path.split(os.sep)) > 1:
        if os.path.splitext(sig_model_path)[-1] == '':
            spec = importlib.util.spec_from_file_location(
                    sig_model_path.split(os.sep)[-1]+'.py',
                    sig_model_path+'.py')
        elif os.path.splitext(sig_model_path)[-1] == '.py':
            spec = importlib.util.spec_from_file_location(
                    sig_model_path.split(os.sep)[-1], sig_model_path)
        else:
            raise argparse.ArgumentTypeError(
                "Specified model file does not end with .py nor does it have "
                "no extension at all. Please specify a valid python file.")
        sig_model = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(sig_model)
    else:
        sig_model = importlib.import_module(
            "pyqmri.models."+str(sig_model_path))
    return sig_model


def _start_recon(myargs):
    sig_model = _import_sigmodel(myargs.sig_model)
# Create par struct to store relevant parameteres for reconstruction/fitting
    par = {}
###############################################################################
# Define precision ############################################################
###############################################################################
    if myargs.double_precision is True:
        par["DTYPE"] = np.complex128
        par["DTYPE_real"] = np.float64
    else:
        par["DTYPE"] = np.complex64
        par["DTYPE_real"] = np.float32
###############################################################################
# Check for 3D input data #####################################################
###############################################################################
    par["is3D"] = myargs.is3Ddata
    if par["is3D"]:
        myargs.use3Dcoilest = True
###############################################################################
# Select input file ###########################################################
###############################################################################
    _readInput(myargs, par)
###############################################################################
# Read Data ###################################################################
###############################################################################
    data, dimX, dimY, NSlice, reco_Slices, dimreduction, off = \
        _read_data_from_file(par, myargs)
###############################################################################
# Flip angle correction #######################################################
###############################################################################
    if "fa_corr" in list(par["file"].keys()) or \
            "interpol_fa" in list(par["file"].keys()):
        print("Using provied flip angle correction.")
        _read_flip_angle_correction_data(
            par,
            myargs,
            dimreduction,
            reco_Slices
            )
    else:
        print("No flip angle correction provided/used.")
###############################################################################
# Check size compatibility ####################################################
###############################################################################
    data = _check_data_size(par, data, myargs.sig_model)
###############################################################################
# Set sequence related parameters #############################################
###############################################################################
    _populate_par_w_sequencepar(par,
                                data.shape,
                                (NSlice, dimY, dimX),
                                myargs.sms,
                                myargs.imagespace,
                                myargs.streamed,
                                myargs.par_slices,
                                myargs.trafo)
    if not myargs.trafo:
        tmpmask = np.ones(data.shape[2:])
        tmpmask[np.abs(data[0, 0, ...]) == 0] = 0
        par['mask'] = np.reshape(
            tmpmask,
            (data.shape[2:])).astype(par["DTYPE_real"])
        del tmpmask
    else:
        par['mask'] = None
###############################################################################
# ratio of z direction to x,y, important for finite differences ###############
###############################################################################
    par["dz"] = par["DTYPE_real"](myargs.dz)
###############################################################################
# Create OpenCL Context and Queues ############################################
###############################################################################
    _setupOCL(myargs, par)
###############################################################################
# Standardize data ############################################################
###############################################################################
    [NScan, NC] = data.shape[:2]

    if myargs.trafo:
        if par["file"].attrs['data_normalized_with_dcf']:
            pass
        else:
            data = data*(par["dcf"])
    if NC == 1:
        par['C'] = np.ones(
            (1, par["NSlice"], dimY, dimX), dtype=par["DTYPE"])
        sumSqrC = np.sqrt(
            np.sum(
                (par["C"] *
                 np.conj(
                    par["C"])),
                0))
        par["C"] = par["C"] / sumSqrC
    else:
###############################################################################
# Coil Sensitivity Estimation #################################################
###############################################################################
        est_coils(data, par, par["file"], myargs, off)
        par['C'] = par['C'].astype(par["DTYPE"])
###############################################################################
# Init forward model and initial guess ########################################
###############################################################################
    if myargs.sig_model == "GeneralModel":
        par["modelfile"] = myargs.modelfile
        par["modelname"] = myargs.modelname
    model = sig_model.Model(par)
###############################################################################
# Reconstruct images using CG-SENSE  ##########################################
###############################################################################
    if myargs.trafo is False:
        data = _precoompFFT(data, par)
    # del par["file"]["images"]
    images = _genImages(myargs, par, data, off)
###############################################################################
# Scale data norm  ############################################################
###############################################################################
    data, images = _estScaleNorm(myargs, par, images, data)
    if np.allclose(myargs.weights, -1):
        if "weights" not in par.keys():
            par["weights"] = np.ones(
                (par["unknowns"]), dtype=par["DTYPE_real"])
    else:
        par["weights"] = np.array(myargs.weights, dtype=par["DTYPE_real"])
###############################################################################
# Compute initial guess #######################################################
###############################################################################
    model.computeInitialGuess(
        images,
        par["dscale"],
        myargs.weights)
###############################################################################
# initialize operator  ########################################################
###############################################################################
    opt = IRGNOptimizer(par,
                        model,
                        trafo=myargs.trafo,
                        imagespace=myargs.imagespace,
                        SMS=myargs.sms,
                        config=myargs.config,
                        streamed=myargs.streamed,
                        reg_type=myargs.reg,
                        DTYPE=par["DTYPE"],
                        DTYPE_real=par["DTYPE_real"])
    f = h5py.File(par["outdir"]+"output_" + par["fname"]+".h5", "a")
    f.create_dataset("images_ifft", data=images)
    f.attrs['data_norm'] = par["dscale"]
    f.close()
    par["file"].close()
###############################################################################
# Start Reco ##################################################################
###############################################################################
    if myargs.imagespace is True:
        opt.execute(images)
    else:
        opt.execute(data)
    plt.close('all')


def _str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    if v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    raise argparse.ArgumentTypeError('Boolean value expected.')


def run(reg_type='TGV',
        slices=1,
        trafo=True,
        streamed=False,
        par_slices=1,
        data='',
        model='GeneralModel',
        config='default',
        imagespace=False,
        sms=False,
        use_GPU=True,
        devices=0,
        dz=1,
        weights=-1,
        useCGguess=True,
        out='',
        modelfile="models.ini",
        modelname="VFA-E1",
        double_precision=False,
        coils3D=False,
        is3Ddata=False):
    """
    Start a 3D model based reconstruction.

    Start a 3D model based reconstruction.
    If no data path is given, a file dialog can be used to select data
    at start up. If no other parameters are passed, T1 from a single slice
    of radially acquired variable flip angle data will be quantified.

    If no config file is passed, a default one will be generated in the
    current folder, the script is run in.

    Parameters
    ----------
      reg_type : str, TGV
        TGV or TV, defaults to TGV
      slices : int, 1
        The number of slices to reconsturct. Slices are picked symmetrically
        from the volume center. Pass -1 to select all slices available.
        Defaults to 1
      trafo : bool, True
        Choos between Radial (1) or Cartesian (0) FFT
      streamed : bool, False
        Toggle between streaming slices to the GPU (1) or computing
        everything with a single memory transfer (0). Defaults to 0
      par_slices : int, 1
        Number of slices per streamed package. Volume devided by GPU's and
        par_slices must be an even number! Defaults to 1
      data : str, ''
        The path to the .h5 file containing the data to reconstruct.
        If left empty, a GUI will open and asks for data file selection. This
        is also the default behaviour.
      model : str, GeneralModel
        The name of the model which should be used to fit the data. Defaults to
        'VFA'. A path to your own model file can be passed. See the Model Class
        for further information on how to setup your own model.
      config : str, default
        The path to the confi gfile used for the IRGN reconstruction. If
        not specified the default config file will be used. If no default
        config file is present in the current working directory one will be
        generated.
      imagespace : bool, False
        Select between fitting in imagespace (1) or in k-space (0).
        Defaults to 0
      sms : bool, False
        use Simultaneous Multi Slice Recon (1) or normal reconstruction (0).
        Defaults to 0
      devices : list of int, 0
        The device ID of device(s) to use for streaming/reconstruction
      dz : float, 1
        Ratio of physical Z to X/Y dimension. X/Y is assumed to be isotropic.
      useCGguess : bool, True
        Switch between CG sense and simple FFT as initial guess for the images.
      out : str, ''
        Output directory. Defaults to the location of the input file.
      modelpath : str, models.ini
        Path to the .mod file for the generative model.
      modelname : str, VFA-E1
        Name of the model in the .mod file to use.
      weights : list of float, -1
        Optional weights for each unknown. Defaults to -1, i.e. no additional
        weights are used.
      double_precision : bool, False
        Enable double precission computation.
    """
    params = [('--recon_type', "TGV"),
              ('--reg_type', str(reg_type)),
              ('--slices', str(slices)),
              ('--trafo', str(trafo)),
              ('--streamed', str(streamed)),
              ('--par_slices', str(par_slices)),
              ('--data', str(data)),
              ('--config', str(config)),
              ('--imagespace', str(imagespace)),
              ('--sms', str(sms)),
              ('--use_GPU', str(use_GPU)),
              ('--devices', str(devices)),
              ('--dz', str(dz)),
              ('--weights', str(weights)),
              ('--useCGguess', str(useCGguess)),
              ('--model', str(model)),
              ('--modelfile', str(modelfile)),
              ('--modelname', str(modelname)),
              ('--out', str(out)),
              ('--double_precision', str(double_precision)),
              ('--estCoils3D', str(coils3D)),
              ('--is3Ddata', str(is3Ddata))
              ]

    sysargs = sys.argv[1:]
    for par_name, par_value in params:
        if par_name not in sysargs:
            sysargs.append(par_name)
            sysargs.append(par_value)
    argsrun, unknown = _parseArguments(sysargs)
    if unknown:
        print("Unknown command line arguments passed: " + str(unknown) + "."
              " These will be ignored for fitting.")
    _start_recon(argsrun)


def _parseArguments(args):
    argparmain = argparse.ArgumentParser(
        description="T1 quantification from VFA "
                    "data. By default runs 3D "
                    "regularization for TGV.")
    argparmain.add_argument(
      '--recon_type', dest='type',
      help='Choose reconstruction type (currently only 3D)')
    argparmain.add_argument(
      '--reg_type', dest='reg',
      help="Choose regularization type (default: TGV) "
           "options are: TGV, TV, all")
    argparmain.add_argument(
      '--slices', dest='slices', type=int,
      help="Number of reconstructed slices (default=40). "
           "Symmetrical around the center slice.")
    argparmain.add_argument(
      '--trafo', dest='trafo', type=_str2bool,
      help='Choos between radial (1, default) and Cartesian (0) sampling. ')
    argparmain.add_argument(
      '--streamed', dest='streamed', type=_str2bool,
      help='Enable streaming of large data arrays (e.g. >10 slices).')
    argparmain.add_argument(
      '--par_slices', dest='par_slices', type=int,
      help='number of slices per package. Volume devided by GPU\'s and'
           ' par_slices must be an even number!')
    argparmain.add_argument(
      '--data', dest='file',
      help="Full path to input data. "
           "If not provided, a file dialog will open.")
    argparmain.add_argument(
      '--config', dest='config',
      help='Name of config file to use (assumed to be in the same folder). \
 If not specified, use default parameters.')
    argparmain.add_argument(
      '--imagespace', dest='imagespace', type=_str2bool,
      help="Select if Reco is performed on images (1) or on kspace (0) data. "
           "Defaults to 0")
    argparmain.add_argument(
      '--sms', dest='sms', type=_str2bool,
      help="Switch to SMS reconstruction")
    argparmain.add_argument(
      '--use_GPU', dest='use_GPU', type=_str2bool,
      help="Select if CPU or GPU should be used as OpenCL platform. "
           "Defaults to GPU (1). CAVE: CPU FFT not working")
    argparmain.add_argument(
      '--devices', dest='devices', type=int,
      help="Device ID of device(s) to use for streaming. "
           "-1 selects all available devices", nargs='*')
    argparmain.add_argument(
      '--dz', dest='dz', type=float,
      help="Ratio of physical Z to X/Y dimension. "
           "X/Y is assumed to be isotropic. Defaults to  1")
    argparmain.add_argument(
      '--weights', dest='weights', type=float,
      help="Ratio of unkowns to each other. Defaults to 1. "
           "If passed, needs to be in the same size as the number of unknowns",
           nargs='*')
    argparmain.add_argument(
      '--useCGguess', dest='usecg', type=_str2bool,
      help="Switch between CG sense and simple FFT as "
           "initial guess for the images.")
    argparmain.add_argument('--out', dest='outdir', type=str,
                            help="Set output directory. Defaults to the input "
                            "file directory")
    argparmain.add_argument(
      '--model', dest='sig_model',
      help='Name of the signal model to use. Defaults to VFA. \
 Please put your signal model file in the Model subfolder.')
    argparmain.add_argument(
      '--modelfile', dest='modelfile', type=str,
      help="Path to the model file.")
    argparmain.add_argument(
      '--modelname', dest='modelname', type=str,
      help="Name of the model to use.")
    argparmain.add_argument(
      '--double_precision', dest='double_precision', type=_str2bool,
      help="Switch between single (False, default) and double "
           "precision (True). Usually, single precision gives high enough "
           "accuracy.")
    argparmain.add_argument(
      '--estCoils3D', dest='use3Dcoilest', type=_str2bool,
      help="Use 3D coil estimation instead of 2D slice-by-slice. \
      Defaults to false. Currently limited to Cartesian data.")
    argparmain.add_argument(
      '--is3Ddata', dest='is3Ddata', type=_str2bool,
      help="Flag to indicate 3D k-space data as input.")

    arguments, unknown = argparmain.parse_known_args(args)
    return arguments, unknown


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Compute Kaiser-Bessel gridding kernel.

Original Skript by B. Hargreaves
Adapted for Python by O. Maier

function [kern,kbu] = calckbkernel(kwidth,overgridfactor,klength)
Function calculates the appropriate Kaiser-Bessel kernel
for gridding, using the approach of Jackson et al.
INPUT:
  width = kernel width in grid samples.
  overgridfactor = over-gridding factor.
  klength = kernel look-up-table length.
OUTPUT:
  kern = kernel values for klength values
  of u, uniformly spaced from 0 to kwidth/2.
  kbu = u values.
"""
import numpy as np
import warnings

def calckbkernel(
        kernelwidth,
        overgridfactor,
        gridsize,
        kernellength=32,
        **kwargs):
    """
    Calculate the appropriate Kaiser-Bessel gridding kernel.
    Function calculates the appropriate Kaiser-Bessel kernel for gridding,
    using the approach of Jackson et al. Original Skript by B. Hargreaves
    Adapted for Python by O. Maier
    Args
    ----
        kernelwidth (int):
            kernel width
        overgridfactor (float):
            over-gridding factor.
        gridsize (int):
            gridsize of oversampled grid
        kernellength (int):
            kernel look-up-table length.
    Returns
    -------
        kern:
            kernel values for kernellength values of u,
            uniformly spaced from 0 to kernelwidth/2.
        kern_ft:
            normalized Fourier transform of kernel used for deapodization
        kbu:
            position of sampling points of kernel.
            linspace from 0 to length of kernel
    """
    if kernellength < 2:
        kernellength = 2
        warnings.warn(
            'Warning:  kernellength must be 2 or more. Default to 2.'
            )

    # From Beatty et al. -
    # Rapid Gridding Reconstruction With a Minimal Oversampling Ratio -
    # equation [5]
    beta = np.pi * np.sqrt(
        (kernelwidth / overgridfactor) ** 2
        * (overgridfactor - 0.5) ** 2
        - 0.8
        )
    # Kernel radii.
    u = np.linspace(
        0,
        (kernelwidth/2),
        int(np.ceil(kernellength*kernelwidth/2)))

    kern = _kb(u, kernelwidth, beta)
    kern = kern / kern[u == 0]  # Normalize.

    ft_y = np.flip(kern)
    if np.mod(kernelwidth, 2):
        ft_y = np.concatenate((ft_y[:-1], kern))
    else:
        ft_y = np.concatenate((ft_y, kern))

    ft_y = np.abs(
      np.fft.fftshift(
          np.fft.ifft(
                  ft_y, gridsize * kernellength
              )
          )
      * ft_y.size
      )

    x = np.linspace(
        -int(np.floor(gridsize/(2*overgridfactor))),
        int(np.floor(gridsize/(2*overgridfactor)))-1,
        int(np.floor(gridsize/(overgridfactor)))
        )

    ft_y = ft_y[(ft_y.size / 2 + x).astype(int)]
    h = np.sinc(x / (gridsize * kernellength)) ** 2

    kern_ft = ft_y * h
    kern_ft = kern_ft / np.max(kern_ft)

    return kern, kern_ft

def _kb(u, w, beta):
    # if np.size(w) > 1:
    #     raise TypeError('w should be a single scalar value.')

    # y = 0 * u  # Allocate space.
    # uz = np.where(np.abs(u) <= w / (2 * G))			# Indices where u<w/2.

    # if np.size(uz) > 0:			# Calculate y at indices uz.
    #     # Argument - see Jackson '91.
    #     x = beta * np.sqrt(1 - (2 * u[uz] * G / w)**2)
    #     y[uz] = G * np.i0(x) / w

    # return y
    """
    Kaiser-Bessel window precomputation.
    Args
    ----
      u (numpy.array):
        Kernel Radii
      width (int):
        Kernel width
      beta (float):
        Scale for the argument of the modified bessel function of oder 0,
        see Jackson '91 and Beatty et al.
    Returns
    -------
      numpy.array
        The Kaiser-Bessel window
    """
    assert np.size(w) == 1, 'width should be a single scalar value.'

    # if np.size(uz) > 0:  # Calculate y at indices uz.
    x = beta * np.sqrt(1 - (2 * u / w) ** 2)
    # Argument - see Jackson '91.
    y = np.i0(x) / w
    return y


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Base for PyOpenCL programs."""
import pyopencl as cl


class CLProgram():
    """Base class for PyOpenCL kernels.

    Parameters
    ----------
      ctx : PyOpenCL.Context
        The context to compile the code in.
      code : string
        The Kernel to compile
    """

    def __init__(self, ctx, code):
        self._cl_prg = cl.Program(ctx, code)
        self._cl_prg.build("-cl-mad-enable -cl-fast-relaxed-math")
        self._cl_kernels = self._cl_prg.all_kernels()
        for kernel in self._cl_kernels:
            self.__dict__[kernel.function_name] = kernel



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Module holding the bi-exponential model for fitting."""
import numpy as np
from pyqmri.models.template import BaseModel, constraints


class Model(BaseModel):
    """Bi-exponential model for MRI parameter quantification.

    This class holds a bi-exponential model for fitting complex MRI data.
    It realizes a forward application of the analytical signal expression
    and the partial derivatives with respesct to each parameter of interest,
    as required by the abstract methods in the BaseModel.

    Parameters
    ----------
      par : dict
        A python dict containing the necessary information to
        setup the object. Needs to contain the sequence related parametrs,
        e.g. TR, TE, TI, to fully describe the acquisitio process

    Attributes
    ----------
      TE : float
        Echo time (or any other timing valid in a bi-exponential fit).
      uk_scale : list of float
        Scaling factors for each unknown to balance the partial derivatives.
      guess : numpy.array, None
        The initial guess. Needs to be set using "computeInitialGuess"
        prior to fitting.
    """

    def __init__(self, par):
        super().__init__(par)
        self.TE = np.ones((self.NScan, 1, 1, 1))

        for i in range(self.NScan):
            self.TE[i, ...] = par["TE"][i] * np.ones((1, 1, 1))

        par["unknowns_TGV"] = 5
        par["unknowns_H1"] = 0
        par["unknowns"] = par["unknowns_TGV"]+par["unknowns_H1"]

        for j in range(par["unknowns"]):
            self.uk_scale.append(1)

        self.constraints.append(
            constraints(
                -100 / self.uk_scale[0],
                100 / self.uk_scale[0], False))
        self.constraints.append(
            constraints(
                0 / self.uk_scale[1],
                1 / self.uk_scale[1],
                True))
        self.constraints.append(
            constraints(
                ((1 / 150) / self.uk_scale[2]),
                ((1 / 1e-4) / self.uk_scale[2]),
                True))
        self.constraints.append(
            constraints(
                0 / self.uk_scale[3],
                1 / self.uk_scale[3],
                True))
        self.constraints.append(
            constraints(
                ((1 / 1500) / self.uk_scale[4]),
                ((1 / 150) / self.uk_scale[4]),
                True))
        self.guess = None

        self._ax = []
        self._M0_plot = None
        self._M0_plot_cor = None
        self._M0_plot_sag = None

        self._M01_plot = None
        self._M01_plot_cor = None
        self._M01_plot_sag = None

        self._M02_plot = None
        self._M02_plot_cor = None
        self._M02_plot_sag = None

        self._T21_plot = None
        self._T21_plot_cor = None
        self._T21_plot_sag = None

        self._T22_plot = None
        self._T22_plot_cor = None
        self._T22_plot_sag = None

    def rescale(self, x):
        """Rescale the unknowns with the scaling factors.

        Rescales each unknown with the corresponding scaling factor and
        applies a 1/x transformation for the time constants of the
        exponentials, yielding a result in milliseconds.

        Parameters
        ----------
          x : numpy.array
            The array of unknowns to be rescaled

        Returns
        -------
          numpy.array:
            The rescaled unknowns
        """
        tmp_x = np.copy(x)
        tmp_x[0] = x[0, ...] * self.uk_scale[0]
        tmp_x[1] = x[1, ...] * self.uk_scale[1]
        tmp_x[2] = 1 / (x[2, ...] * self.uk_scale[2])
        tmp_x[3] = x[3, ...] * self.uk_scale[3]
        tmp_x[4] = 1 / (x[4, ...] * self.uk_scale[4])

        const = []
        for constrained in self.constraints:
            const.append(constrained.real)
        return {"data": tmp_x,
                "unknown_name": ["M0", "M01", "T21",
                                 "M02", "T22"],
                "real_valued": const}

    def _execute_forward_3D(self, x):
        M0 = x[0, ...] * self.uk_scale[0]
        M01 = x[1, ...] * self.uk_scale[1]
        T21 = x[2, ...] * self.uk_scale[2]
        M02 = x[3, ...] * self.uk_scale[3]
        T22 = x[4, ...] * self.uk_scale[4]
        S = M0 * (M01 * np.exp(-self.TE * (T21)) +
                  M02 * np.exp(-self.TE * (T22)))
        S[~np.isfinite(S)] = 1e-20
        S = np.array(S, dtype=self._DTYPE)
        return S

    def _execute_gradient_3D(self, x):
        M0 = x[0, ...]
        M01 = x[1, ...]
        T21 = x[2, ...]
        M02 = x[3, ...]
        T22 = x[4, ...]
        grad_M0 = self.uk_scale[0] * (
            M01 * self.uk_scale[1] * np.exp(-self.TE * (
                T21 * self.uk_scale[2])) +
            M02 * self.uk_scale[3] *
            np.exp(-self.TE * (T22 * self.uk_scale[4])))
        grad_M01 = self.uk_scale[0] * M0 * self.uk_scale[1] * \
            np.exp(-self.TE * (T21 * self.uk_scale[2]))
        grad_T21 = -self.uk_scale[0] * M0 * M01 * self.uk_scale[1] * \
            self.TE * \
            self.uk_scale[2] * np.exp(-self.TE * (T21 * self.uk_scale[2]))
        grad_M02 = self.uk_scale[0] * M0 * self.uk_scale[3] * \
            np.exp(-self.TE * (T22 * self.uk_scale[4]))
        grad_T22 = -self.uk_scale[0] * M0 * M02 * self.uk_scale[3] * \
            self.TE * \
            self.uk_scale[4] * np.exp(-self.TE * (T22 * self.uk_scale[4]))
        grad = np.array([grad_M0, grad_M01, grad_T21,
                         grad_M02, grad_T22], dtype=self._DTYPE)
        grad[~np.isfinite(grad)] = 1e-20
        return grad

    def computeInitialGuess(self, *args):
        """Initialize unknown array for the fitting.

        This function provides an initial guess for the fitting.

        Parameters
        ----------
          args : list of objects
            Serves as universal interface. No objects need to be passed
            here.
        """
        test_M0 = 0.1 * np.ones(
            (self.NSlice, self.dimY, self.dimX),
            dtype=self._DTYPE)
        test_M01 = 0.5 * np.ones(
            (self.NSlice, self.dimY, self.dimX), dtype=self._DTYPE)
        test_T11 = 1/10 * np.ones(
            (self.NSlice, self.dimY, self.dimX), dtype=self._DTYPE)
        test_T12 = 1/150 * np.ones(
            (self.NSlice, self.dimY, self.dimX), dtype=self._DTYPE)

        self.guess = np.array([
            test_M0,
            test_M01,
            test_T11,
            test_M01,
            test_T12], dtype=self._DTYPE)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Simple Volumetric image viewer with scrolling option."""
import matplotlib.pyplot as plt
import numpy as np

###############################################################################
# 3D viewer  ##########s#######################################################
###############################################################################


def imshow(volume, vmin=None, vmax=None):
    """Volumetric image viewer for python.

    Shows up to 4D volumes. volume is assumed to be real valued.

    Parameters
    ----------
      volume : numpy.array
        The (real values) image data to display
      vmin : float, None
        Minimum value of the display window. If None, uses minimum value in
        middle slice of the volume.
      vmax : float, None
        Maximum value of the display window. If None, uses maximum value in
        middle slice of the volume.
    """
    plot_rgb = 0
    if volume.shape[-1] == 3:
        plot_rgb = 1
    if volume.ndim - plot_rgb == 2:
        fig, ax = plt.subplots()
        ax.volume = volume
        ax.index = None
        ax.imshow(volume, vmin=vmin, vmax=vmax)
    elif volume.ndim - plot_rgb == 3:
        fig, ax = plt.subplots()
        ax.volume = volume
        ax.index = volume.shape[0] // 2
        ax.imshow(volume[ax.index], vmin=vmin, vmax=vmax)
    elif volume.ndim - plot_rgb == 4:
        fig, ax = plt.subplots(int(np.ceil(np.sqrt(volume.shape[0]))), int(
            np.ceil(np.sqrt(volume.shape[0]))))
        fig.tight_layout()
        fig.subplots_adjust(hspace=0, wspace=0)
        ax = ax.flatten()
        ni = int(np.ceil(np.sqrt(volume.shape[0])))
        nj = int(np.ceil(np.sqrt(volume.shape[0])))
        for j in range(nj):
            for i in range(ni):
                ax[i + ni * j].axis('off')
                if i + ni * j >= volume.shape[0]:
                    ax[i + ni * j].volume = np.zeros_like(volume[0])
                else:
                    ax[i + ni * j].volume = volume[i + (j * ni)]
                    ax[i + ni * j].index = volume[i + (j * ni)].shape[0] // 2
                    ax[i + ni * j].imshow(volume[i + (j * ni),
                                                 ax[i + ni * j].index],
                                          vmin=vmin, vmax=vmax)
    else:
        raise NameError('Unsupported Dimensions')
    fig.canvas.mpl_connect('scroll_event', _process_scroll)


def _process_scroll(event):
    fig = event.canvas.figure
    ax = fig.axes
    for i, axes in enumerate(ax):
        if axes.index is not None:
            volume = axes.volume
            if (int((axes.index - event.step) >= volume.shape[0]) or
                    int((axes.index - event.step) < 0)):
                pass
            else:
                ax[i].index = int((axes.index - event.step) % volume.shape[0])
                ax[i].images[0].set_array(volume[ax[i].index])
                fig.canvas.draw()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Non-linear Inversion.

nlinvns
 Written and invented
 by Martin Uecker <muecker@gwdg.de> in 2008-09-22

 Modifications by Tilman Sumpf 2012 <tsumpf@gwdg.de>:
     removed fftshift during reconstruction (ns = "no shift")
     added switch to return coil profiles
     added switch to force the image estimate to be real
     use of vectorized operation rather than "for" loops

 Version 0.1

 Biomedizinische NMR Forschungs GmbH am
 Max-Planck-Institut fuer biophysikalische Chemie
Adapted for Python by O. Maier
"""
from time import perf_counter
import numpy as np

from pyqmri.transforms import PyOpenCLnuFFT
import pyopencl.array as cla


def nlinvns(Y, n, par, *arg, DTYPE=np.complex64, DTYPE_real=np.float32):
    """Non-linear inversen based Coil sensitivity estimation.

    Parameters
    ----------
      Y : numpy.array
        Data used for estimation
      n : int
        number of Gausse-Newton iteration steps
      realConstr : bool
        Real value constraint on the image. Should be set to false usually.
      returnProfiles : bool
        Return coil profiles. Should be set to True usually.
      DTYPE : numpy.dtype, numpy.complex64
        Complex working precission.
      DTYPE_real : numpy.dtype, numpy.float32
        Real working precission.

    Returns
    -------
    numpy.array :
        The results of each iteration containing the reconstructed image at
        position 0 and 1, followed by the complex coil sensitivities.
    """
    nrarg = len(arg)
    if nrarg == 2:
        returnProfiles = arg[0]
        realConstr = arg[1]
    elif nrarg < 2:
        realConstr = False
        if nrarg < 1:
            returnProfiles = 0

    print('Start...')

    alpha = 0.001

    [c, z, y, x] = Y.shape
    
    par["NC"] = c
    par["fft_dim"] = [-3, -2, -1]
    par["mask"] = np.ones(Y.shape[1:], dtype=par["DTYPE_real"])
    
    FFT = PyOpenCLnuFFT.create(
        par["ctx"][0], par["queue"][0], par,
        DTYPE=par["DTYPE"],
        DTYPE_real=par["DTYPE_real"],
        radial=False, SMS=False)
       

    if returnProfiles:
        R = np.zeros([c + 2, n, z, y, x], DTYPE)

    else:
        R = np.zeros([2, n, z, y, x], DTYPE)

    # initialization x-vector
    X0 = np.array(np.zeros([c + 1, z, y, x]), DTYPE)
    X0[0] = 1  # object part

    # initialize mask and weights
    P = np.ones(Y[0].shape, dtype=DTYPE)
    P[Y[0] == 0] = 0

    W = _weights(z, y, x)

    W = _fftshift2(W).astype(DTYPE_real)

    # normalize data vector
    yscale = 100 / np.sqrt(_scal(Y, Y))
    YS = Y * yscale

    XT = np.zeros([c + 1, z, y, x], dtype=DTYPE)
    XN = np.copy(X0)
    
    start = perf_counter()  
       
    for i in range(0, n):

        # the application of the weights matrix to XN
        # is moved out of the operator and the derivative
        XT[0] = np.copy(XN[0])
        XT[1:] = _apweightsns(W, np.copy(XN[1:]), FFT)

        RES = (YS - _opns(P, XT, FFT))

        print(np.round(np.linalg.norm(RES)))

        # calculate rhs
        r = _derHns(P, W, XT, RES, realConstr, FFT)

        r = np.array(r + alpha * (X0 - XN), dtype=DTYPE)

        z = np.zeros_like(r)
        d = np.copy(r)
        dnew = np.linalg.norm(r)**2
        dnot = np.copy(dnew)

        for j in range(0, 100):

            # regularized normal equations
            q = _derHns(P, W, XT, _derns(P, W, XT, d, FFT), realConstr, FFT) + alpha * d
            # np.nan_to_num(q)

            a = dnew / np.real(_scal(d, q))
            z = z + a * d
            r = r - a * q
            # np.nan_to_num(r)
            dold = np.copy(dnew)
            dnew = np.linalg.norm(r)**2

            d = d * ((dnew / dold)) + r
            # np.nan_to_num(d)
            if np.sqrt(dnew) < (1e-4 * dnot):
                break

        print('(', j, ')')

        XN = XN + z

        alpha = alpha / 3

        # postprocessing

        CR = _apweightsns(W, XN[1:, :, :], FFT)

        if returnProfiles:
            R[2:, i] = CR / yscale

        C = (np.conj(CR) * CR).sum(0)

        R[0, i] = (XN[0] * np.sqrt(C) / yscale)
        R[1, i] = np.copy(XN[0])

    end = perf_counter()
    print('done in', round((end - start)), 's')
    return R


def _scal(a, b):
    v = np.sum(np.conj(a) * b)
    return v


def _apweightsns(W, CT, fft):
    C = _nsIfft(W * CT, fft)
    return C


def _apweightsnsH(W, CT, fft):
    C = np.conj(W) * _nsFft(CT, fft)
    return C


def _opns(P, X, fft):
    K = X[0] * X[1:]
    K = P * _nsFft(K, fft)
    return K


def _derns(P, W, X0, DX, fft):
    K = X0[0] * _apweightsns(W, DX[1:], fft)
    K = K + (DX[0] * X0[1:])
    K = P * _nsFft(K, fft)
    return K


def _derHns(P, W, X0, DK, realConstr, fft):

    K = _nsIfft(P * DK, fft)

    if realConstr:
        DXrho = np.sum(np.real(K * np.conj(X0[1:])), 0)
    else:
        DXrho = np.sum(K * np.conj(X0[1:]), 0)

    DXc = _apweightsnsH(W, (K * np.conj(X0[0])), fft)
    DX = np.concatenate(
        (DXrho[None, ...], DXc), axis=0)
    return DX


def _nsFft(M, fft):
    # K = np.fft.fftn(M, axes=(-3,-2,-1), norm="ortho")
    tmp = cla.to_device(fft.queue, M)
    K = cla.empty_like(tmp)
    fft.FFT(K,tmp)
    # K = fft(M)
    return K.get()


def _nsIfft(M, fft):
    # K = np.fft.ifftn(M, axes=(-3,-2,-1), norm="ortho")
    tmp = cla.to_device(fft.queue, M)
    K = cla.empty_like(tmp)
    fft.FFTH(K,tmp)
    # K = fft(M)
    return K.get()


def _weights(z, y, x):
    W = np.zeros([z, y, x])
    for k in range(z):
        for i in range(y):
            for j in range(x):
                d = ((i) / y - 0.5)**2 + ((j) / x - 0.5)**2 + ((k) / z - 0.5)**2
                W[k, i, j] = 1 / (1 + 220 * d)**16
    return W


def _fftshift2(I):
    if I.ndim >= 3:
        S = np.fft.fftshift(I, axes=(-3,-2,-1))
    else:
        S = np.fft.fftshift(np.fft.fftshift(I, 1), 0)
    return S



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Module holding the intravoxel incoherent motion  (IVIM) model for fitting."""
from pyqmri.models.template import BaseModel, constraints
import numpy as np


class Model(BaseModel):
    """Intravoxel incoherent motion (IVIM) model for MRI parameter quantification.
    
    This class holds an IVIM model for fitting complex MRI data.
    It realizes a forward application of the analytical signal expression
    and the partial derivatives with respect to each parameter of interest,
    as requierd by the abstract methods in the BaseModel.
    
    The fitting is based on the two-compartment bi-exponential IVIM approach, 
    assuming a pure tissue diffusion and a pseudo-diffusion compartment
    related to perfusion.
    
    Parameters
    ----------
      par : dict
        A python dict containing the necessary information to
        setup the object. Needs to contain the sequence related parameters, 
        e.g. b-values, to fully describe the acquisition process
        
    Attributes
    ----------
      b : float
        b values for each diffusino direction.
      dir : numpy.array
        The diffusion direction vectors. Assumed to have length 1.
      uk_scale : list of float
        Scaling factors for each unknown to balance the partial derivatives.
      guess : numpy.array
        The initial guess. Needs to be set using "computeInitialGuess"
        prior to fitting.
      phase : numpy.array
        The phase of each diffusion direction relative to the b0 image.
        Estimated during the initial guess using the image series of all
        directions/bvalue pairs.
      b0 : numpy.array
        The b0 image if present in the data file. None else.
    """  
           
    def __init__(self, par):
        super().__init__(par)
        self.NSlice = par['NSlice']

        self.figure_phase = None

        self.b = np.ones((self.NScan, 1, 1, 1))
        self.dir = par["DWI_dir"].T
        for i in range(self.NScan):
            self.b[i, ...] = par["b_value"][i] * np.ones((1, 1, 1))

        if np.max(self.b) > 100:
            self.b /= 1000

        self.dir = self.dir[:, None, None, None, :]
        par["unknowns_TGV"] = 4
        par["unknowns_H1"] = 0
        par["unknowns"] = par["unknowns_TGV"] + par["unknowns_H1"]
        self.unknowns = par["unknowns_TGV"] + par["unknowns_H1"]
        self.uk_scale = []
        for j in range(self.unknowns):
            self.uk_scale.append(1)

        try:
            self.b0 = np.flip(
                np.transpose(par["file"]["b0"][()], (0, 2, 1)), 0)
        except KeyError:
            print("No b0 image provided")
            self.b0 = None

        self.constraints.append(
            constraints(
                0 / self.uk_scale[0],
                10 / self.uk_scale[0],
                False))
        self.constraints.append(
            constraints(
                (0 / self.uk_scale[1]),
                (5 / self.uk_scale[1]),
                True))
        self.constraints.append(
            constraints(
                (0.01 / self.uk_scale[2]),
                (1 / self.uk_scale[2]),
                True))
        self.constraints.append(
            constraints(
                (5 / self.uk_scale[3]),
                (300 / self.uk_scale[3]),
                True))
        self.ivim_scale = 1
        
    def rescale(self, x):
        tmp_x = np.copy(x)
        tmp_x[0] = x[0, ...] * self.uk_scale[0]
        tmp_x[1] = x[1, ...] * self.uk_scale[1]
        tmp_x[2] = x[2, ...] * self.uk_scale[2]
        tmp_x[3] = x[3, ...] * self.uk_scale[3]*self.ivim_scale
        
        const = []
        for constrained in self.constraints:
            const.append(constrained.real)
        return{"data":tmp_x, 
               "unknown_name": ["M0", "ADC", "f", "ADC_ivim"],
               "real_valued": const}

    def _execute_forward_2D(self, x, islice):
        print("2D Functions not implemented")
        raise NotImplementedError

    def _execute_gradient_2D(self, x, islice):
        print("2D Functions not implemented")
        raise NotImplementedError

    def _execute_forward_3D(self, x):        
        ADC = x[1, ...] * self.uk_scale[1]

        S = (x[0, ...] * self.uk_scale[0] * (
                x[2, ...] * self.uk_scale[2]
                * np.exp(-(x[3, ...] * self.uk_scale[3]) * self.b * self.ivim_scale)
                + (1-x[2, ...] * self.uk_scale[2])
                * np.exp(- ADC * self.b)
             )).astype(self._DTYPE)

        S *= self.phase
        S[~np.isfinite(S)] = 0
        return S

    def _execute_gradient_3D(self, x):
        ADC = x[1, ...] * self.uk_scale[1]

        grad_M0 = self.uk_scale[0] * (
            x[2, ...] * self.uk_scale[2]
            * np.exp(- (x[3, ...] * self.uk_scale[3]) * self.b * self.ivim_scale)
            + (1-x[2, ...] * self.uk_scale[2])
            * np.exp(- ADC * self.b))

        grad_ADC = x[0, ...] * self.uk_scale[0] * (
            - self.b * self.uk_scale[1] * np.exp(- ADC * self.b) + (
                x[2, ...] * self.b * self.uk_scale[2] * self.uk_scale[1] * np.exp(- ADC * self.b)))
        

        grad_f = (x[0, ...] * self.uk_scale[0] * self.uk_scale[2] * (
            np.exp(-(x[3, ...] * self.uk_scale[3]) * self.b * self.ivim_scale)
            - np.exp(- ADC * self.b)))

        grad_ADC_ivim = (
            -x[0, ...] * self.b*self.uk_scale[0]*self.ivim_scale * self.uk_scale[3] * (
                x[2, ...] * self.uk_scale[2] *
                np.exp(- (x[3, ...] * self.uk_scale[3]) * self.b * self.ivim_scale))
            )

        grad = np.array(
            [grad_M0,
             grad_ADC,
             grad_f,
             grad_ADC_ivim], dtype=self._DTYPE)
        grad[~np.isfinite(grad)] = 0
        grad *= self.phase
        return grad


    def computeInitialGuess(self, *args):
        """Initialize unknown array for the fiting.
        This function provides an initial guess for the fiting. args[0] is
        assumed to contain the image series which is used for phase 
        correction.
        
        Parameters
        ----------
          args : list of objects
            Assumes the image series at position 0 and optionally computes
            a phase based on the difference between each image series minus
            the first image in the series (Scan i minus Scan 0). This 
            phase correction is needed as each diffusion weighting has a 
            different phase.
        """
        
        self.phase = np.exp(1j*(np.angle(args[0])-np.angle(args[0][0])))
        if self.b0 is not None:
            test_M0 = self.b0
        else:
            test_M0 = args[0][0]
            

        if np.allclose(args[2],-1):  
            # default setting
            ADC = 1 * np.ones(args[0].shape[-3:], dtype=self._DTYPE)
            f = 0.2 * np.ones(args[0].shape[-3:], dtype=self._DTYPE)
            ADC_ivim = 20 * np.ones(args[0].shape[-3:], dtype=self._DTYPE)
        else:
            assert len(args[2]) == self.unknowns-1

            ADC  = args[2][0] *np.ones(args[0].shape[-3:], dtype=self._DTYPE)
            f = args[2][1] * np.ones(args[0].shape[-3:], dtype=self._DTYPE)
            ADC_ivim = args[2][2] * np.ones(args[0].shape[-3:], dtype=self._DTYPE)
            

        x = np.array(
                [
                    test_M0 / self.uk_scale[0],
                    ADC,
                    f,
                    ADC_ivim],
                dtype=self._DTYPE)
        x_scale = np.max(np.abs(x).reshape(x.shape[0], -1), axis=-1)
        self.uk_scale = x_scale
        self.guess = x/x_scale[:,None,None,None]
        for uk in range(self.unknowns):
            self.constraints[uk].update(x_scale[uk])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Module holding the classes for different FFT operators."""
import numpy as np
import pyopencl as cl
import pyopencl.array as clarray
from gpyfft.fft import FFT
from pkg_resources import resource_filename
from pyqmri._helper_fun._calckbkernel import calckbkernel
from pyqmri._helper_fun import CLProgram as Program


class PyOpenCLnuFFT():
    """Base class for FFT calculation.

    This class serves as the base class for all FFT object used in
    the varous optimization algorithms. It provides a factory method
    to generate a FFT object based on the input.

    Parameters
    ----------
        ctx : PyOpenCL.Context
          The context for the PyOpenCL computations.
        queue : PyOpenCL.Queue
          The computation Queue for the PyOpenCL kernels.
        fft_dim : tuple of int
          The dimensions to take the fft over
        DTYPE : Numpy.dtype
          The comlex precision type. Currently complex64 is used.
        DTYPE_real : Numpy.dtype
          The real precision type. Currently float32 is used.

    Attributes
    ----------
      DTYPE : Numpy.dtype
        The comlex precision type. Currently complex64 is used.
      DTYPE_real : Numpy.dtype
        The real precision type. Currently float32 is used.
      ctx : PyOpenCL.Context
        The context for the PyOpenCL computations.
      queue : PyOpenCL.Queue
        The computation Queue for the PyOpenCL kernels.
      prg : PyOpenCL.Program
        The PyOpenCL Program Object containing the compiled kernels.
      fft_dim : tuple of int
        The dimensions to take the fft over
    """

    def __init__(self, ctx, queue, fft_dim, DTYPE, DTYPE_real):
        self.DTYPE = DTYPE
        self.DTYPE_real = DTYPE_real
        self.ctx = ctx
        self.queue = queue
        self.prg = None
        self.fft_dim = fft_dim

    @staticmethod
    def create(ctx,
               queue,
               par,
               kwidth=5,
               klength=1000,
               DTYPE=np.complex64,
               DTYPE_real=np.float32,
               radial=False,
               SMS=False,
               streamed=False):
        """FFT factory method.

        Based on the inputs this method decides which FFT object should be
        returned.

        Parameters
        ----------
          ctx : PyOpenCL.Context
            The context for the PyOpenCL computations.
          queue : PyOpenCL.Queue
            The computation Queue for the PyOpenCL kernels.
          par : dict
            A python dict containing the necessary information to setup the
            object. Needs to contain the number of slices (NSlice), number of
            scans (NScan), image dimensions (dimX, dimY), number of coils (NC),
            sampling points (N) and read outs (NProj) a PyOpenCL queue (queue)
            and the complex coil sensitivities (C).
          kwidth : int, 5
            The width of the sampling kernel for regridding of non-uniform
            kspace samples.
          klength : int, 200
            The length of the kernel lookup table which samples the contineous
            gridding kernel.
          DTYPE : Numpy.dtype, numpy.complex64
            The comlex precision type. Currently complex64 is used.
          DTYPE_real : Numpy.dtype, numpy.float32
            The real precision type. Currently float32 is used.
          radial : bool, False
            Switch for Cartesian (False) and non-Cartesian (True) FFT.
          SMS : bool, False
            Switch between Simultaneous Multi Slice reconstruction (True) and
            simple slice by slice reconstruction.
          streamed : bool, False
            Switch between normal reconstruction in one big block versus
            streamed reconstruction of smaller blocks.

        Returns
        -------
          PyOpenCLnuFFT object:
            The setup FFT object.

        Raises
        ------
          AssertionError:
            If the Combination of passed flags to choose the
            FFT aren't compatible with each other. E.g.: Radial and SMS True.
        """
        if not streamed:
            if radial is True and SMS is False:
                if par["is3D"]:
                    obj = PyOpenCL3DRadialNUFFT(
                        ctx,
                        queue,
                        par,
                        kwidth=kwidth,
                        klength=klength,
                        DTYPE=DTYPE,
                        DTYPE_real=DTYPE_real)
                else:
                    obj = PyOpenCLRadialNUFFT(
                        ctx,
                        queue,
                        par,
                        kwidth=kwidth,
                        klength=klength,
                        DTYPE=DTYPE,
                        DTYPE_real=DTYPE_real)
            elif SMS is True and radial is False:
                obj = PyOpenCLSMSNUFFT(
                    ctx,
                    queue,
                    par,
                    DTYPE=DTYPE,
                    DTYPE_real=DTYPE_real)
            elif SMS is False and radial is False:
                obj = PyOpenCLCartNUFFT(
                    ctx,
                    queue,
                    par,
                    DTYPE=DTYPE,
                    DTYPE_real=DTYPE_real)
            else:
                raise AssertionError("Combination of Radial "
                                     "and SMS not allowed")
            if DTYPE == np.complex128:
                print('Using double precision')
                file = open(
                    resource_filename(
                        'pyqmri', 'kernels/OpenCL_gridding_double.c'))
                obj.prg = Program(
                    obj.ctx,
                    file.read())
            else:
                print('Using single precision')
                file = open(
                    resource_filename(
                        'pyqmri', 'kernels/OpenCL_gridding_single.c'))
                obj.prg = Program(
                    obj.ctx,
                    file.read())
        else:
            if radial is True and SMS is False:
                if par["is3D"]:
                    raise NotImplementedError("3D non-cartesian and streamed\
                                              not implemented")
                obj = PyOpenCLRadialNUFFT(
                    ctx,
                    queue,
                    par,
                    kwidth=kwidth,
                    klength=klength,
                    DTYPE=DTYPE,
                    DTYPE_real=DTYPE_real,
                    streamed=True)
            elif SMS is True and radial is False:
                obj = PyOpenCLSMSNUFFT(
                    ctx,
                    queue,
                    par,
                    DTYPE=DTYPE,
                    DTYPE_real=DTYPE_real,
                    streamed=True)
            elif SMS is False and radial is False:
                obj = PyOpenCLCartNUFFT(
                    ctx,
                    queue,
                    par,
                    DTYPE=DTYPE,
                    DTYPE_real=DTYPE_real,
                    streamed=True)
            else:
                raise AssertionError("Combination of Radial "
                                     "and SMS not allowed")
            if DTYPE == np.complex128:
                print('Using double precision')
                file = open(
                    resource_filename(
                        'pyqmri',
                        'kernels/OpenCL_gridding_slicefirst_double.c'))
                obj.prg = Program(
                    obj.ctx,
                    file.read())
            else:
                print('Using single precision')
                file = open(
                    resource_filename(
                        'pyqmri',
                        'kernels/OpenCL_gridding_slicefirst_single.c'))
                obj.prg = Program(
                    obj.ctx,
                    file.read())
        file.close()
        return obj


class PyOpenCLRadialNUFFT(PyOpenCLnuFFT):
    """Non-uniform FFT object.

    This class performs the non-uniform FFT (NUFFT) operation. Linear
    interpolation of a sampled gridding kernel is used to regrid points
    from the non-cartesian grid back on the cartesian grid.

    Parameters
    ----------
      ctx : PyOpenCL.Context
        The context for the PyOpenCL computations.
      queue : PyOpenCL.Queue
        The computation Queue for the PyOpenCL kernels.
      par : dict
        A python dict containing the necessary information to
        setup the object. Needs to contain the number of slices (NSlice),
        number of scans (NScan), image dimensions (dimX, dimY), number of
        coils (NC), sampling points (N) and read outs (NProj)
        a PyOpenCL queue (queue) and the complex coil
        sensitivities (C).
      kwidth : int
        The width of the sampling kernel for regridding of non-uniform
        kspace samples.
      klength : int
        The length of the kernel lookup table which samples the contineous
        gridding kernel.
      DTYPE : Numpy.dtype
        The comlex precision type. Currently complex64 is used.
      DTYPE_real : Numpy.dtype
        The real precision type. Currently float32 is used.

    Attributes
    ----------
      traj : PyOpenCL.Array
        The comlex sampling trajectory
      dcf : PyOpenCL.Array
        The densitiy compenation function
      ogf (float):
        The overgriddingfactor for non-cartesian k-spaces.
      fft_shape : tuple of ints
        3 dimensional tuple. Dim 0 containts all Scans, Coils and Slices.
        Dim 1 and 2 the overgridded image dimensions.
      fft_scale : float32
        The scaling factor to achieve a good adjointness of the forward and
        backward FFT.
      cl_kerneltable (PyOpenCL.Buffer):
        The gridding lookup table as read only Buffer
      cl_deapo (PyOpenCL.Buffer):
        The deapodization lookup table as read only Buffer
      par_fft : int
        The number of parallel fft calls. Typically it iterates over the
        Scans.
      fft : gpyfft.fft.FFT
        The fft object created from gpyfft (A wrapper for clFFT). The object
        is created only once an reused in each iterations, iterationg over
        all scans to keep the memory footprint low.
      prg : PyOpenCL.Program
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator. This will be determined by the
        factory and set after the object is created.
    """

    def __init__(
            self,
            ctx,
            queue,
            par,
            kwidth=5,
            klength=200,
            DTYPE=np.complex64,
            DTYPE_real=np.float32,
            streamed=False):
        super().__init__(ctx, queue, par["fft_dim"], DTYPE, DTYPE_real)
        
        self.ogf = par["ogf"]
        
        if streamed:
            self.fft_shape = (
                par["NScan"] *
                par["NC"] *
                (par["par_slices"] + par["overlap"]),
                int(round(par["dimY"]*self.ogf)),
                int(round(par["dimX"]*self.ogf)))
        else:
            self.fft_shape = (
                par["NScan"] *
                par["NC"] *
                par["NSlice"],
                int(round(par["dimY"]*self.ogf)),
                int(round(par["dimX"]*self.ogf)))
        
        self.fft_scale = DTYPE_real(
            np.sqrt(np.prod(self.fft_shape[self.fft_dim[0]:])))

        (kerneltable, kerneltable_FT) = calckbkernel(
            kwidth, self.ogf, par["N"], klength)

        deapo = 1 / kerneltable_FT.astype(DTYPE_real)

        self.cl_kerneltable = cl.Buffer(
            self.ctx,
            cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
            hostbuf=kerneltable.astype(DTYPE_real).data)
        self.cl_deapo = cl.Buffer(
            self.ctx,
            cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
            hostbuf=deapo.data)
        self.dcf = clarray.to_device(self.queue, par["dcf"])
        self.traj = clarray.to_device(self.queue, par["traj"])
        self._tmp_fft_array = (
            clarray.zeros(
                self.queue,
                (self.fft_shape),
                dtype=DTYPE))
        if par["use_GPU"]:
            self.par_fft = int(
                self.fft_shape[0] / par["NScan"])
        else:
            self.par_fft = self.fft_shape[0]
        self.iternumber = int(self.fft_shape[0]/self.par_fft)
        self.fft = FFT(ctx, queue, self._tmp_fft_array[
            0:self.par_fft, ...],
                       out_array=self._tmp_fft_array[
                           0:self.par_fft, ...],
                       axes=self.fft_dim)

        self._kernelpoints = kerneltable.size
        self._kwidth = kwidth / 2
        self._check = np.ones(self.fft_shape[-1], dtype=DTYPE_real)
        self._check[1::2] = -1
        self._check = clarray.to_device(self.queue, self._check)
        self._gridsize = self.fft_shape[-1]

    def __del__(self):
        """Explicitly delete OpenCL Objets."""
        del self.traj
        del self.dcf
        del self._tmp_fft_array
        del self.cl_kerneltable
        del self.cl_deapo
        del self._check
        del self.queue
        del self.ctx
        del self.prg
        del self.fft

    def FFTH(self, sg, s, wait_for=None, scan_offset=0):
        """Perform the inverse (adjoint) NUFFT operation.

        Parameters
        ----------
          sg : PyOpenCL.Array
            The complex image data.
          s : PyOpenCL.Array
            The non-uniformly gridded k-space
          wait_for : list of PyopenCL.Event, None
            A List of PyOpenCL events to wait for.
          scan_offset : int, 0
            Offset compared to the first acquired scan.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        if wait_for is None:
            wait_for = []
        # Zero tmp arrays
        self._tmp_fft_array.add_event(
            self.prg.zero_tmp(
                self.queue,
                (self._tmp_fft_array.size,
                 ),
                None,
                self._tmp_fft_array.data,
                wait_for=self._tmp_fft_array.events))
        # Grid k-space
        self._tmp_fft_array.add_event(
            self.prg.grid_lut(
                self.queue,
                (s.shape[0], s.shape[1] * s.shape[2],
                 s.shape[-2] * self._gridsize),
                None,
                self._tmp_fft_array.data,
                s.data,
                self.traj.data,
                np.int32(self._gridsize),
                np.int32(sg.shape[2]),
                self.DTYPE_real(self._kwidth),
                self.dcf.data,
                self.cl_kerneltable,
                np.int32(self._kernelpoints),
                np.int32(scan_offset),
                wait_for=(wait_for + 
                          s.events + self._tmp_fft_array.events)))
        # FFT
        self._tmp_fft_array.add_event(
            self.prg.fftshift(
                self.queue,
                (self.fft_shape[0],
                 self.fft_shape[1],
                 self.fft_shape[2]),
                None,
                self._tmp_fft_array.data,
                self._check.data,
                wait_for=self._tmp_fft_array.events))  
        cl.wait_for_events(self._tmp_fft_array.events)
        fft_events = []
        for j in range(self.iternumber):
                fft_events.append(self.fft.enqueue_arrays(
                    data=self._tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    result=self._tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    forward=False)[0])
        self._tmp_fft_array.add_event(
            self.prg.fftshift(
                self.queue,
                (self.fft_shape[0],
                 self.fft_shape[1],
                 self.fft_shape[2]),
                None,
                self._tmp_fft_array.data,
                self._check.data,
                wait_for=fft_events))
        return self.prg.deapo_adj(
            self.queue,
            (sg.shape[0] * sg.shape[1] *
             sg.shape[2], sg.shape[3], sg.shape[4]),
            None,
            sg.data,
            self._tmp_fft_array.data,
            self.cl_deapo,
            np.int32(self._tmp_fft_array.shape[-1]),
            self.DTYPE_real(self.fft_scale),
            self.DTYPE_real(self.ogf),
            wait_for=(wait_for + sg.events + 
                      self._tmp_fft_array.events))

    def FFT(self, s, sg, wait_for=None, scan_offset=0):
        """Perform the forward NUFFT operation.

        Parameters
        ----------
          s : PyOpenCL.Array
            The non-uniformly gridded k-space.
          sg : PyOpenCL.Array
            The complex image data.
          wait_for : list of PyopenCL.Event, None
            A List of PyOpenCL events to wait for.
          scan_offset : int, 0
            Offset compared to the first acquired scan.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        if wait_for is None:
            wait_for = []
        # Zero tmp arrays
        self._tmp_fft_array.add_event(
            self.prg.zero_tmp(
                self.queue,
                (self._tmp_fft_array.size,
                 ),
                None,
                self._tmp_fft_array.data,
                wait_for=
                self._tmp_fft_array.events))
        # Deapodization and Scaling
        self._tmp_fft_array.add_event(
            self.prg.deapo_fwd(
                self.queue,
                (sg.shape[0] * sg.shape[1] * sg.shape[2],
                 sg.shape[3], sg.shape[4]),
                None,
                self._tmp_fft_array.data,
                sg.data,
                self.cl_deapo,
                np.int32(self._tmp_fft_array.shape[-1]),
                self.DTYPE_real(1 / self.fft_scale),
                self.DTYPE_real(self.ogf),
                wait_for=wait_for + sg.events + self._tmp_fft_array.events))
        # FFT
        self._tmp_fft_array.add_event(
            self.prg.fftshift(
                self.queue,
                (self.fft_shape[0],
                 self.fft_shape[1],
                 self.fft_shape[2]),
                None,
                self._tmp_fft_array.data,
                self._check.data,
                wait_for=self._tmp_fft_array.events))
        
        cl.wait_for_events(self._tmp_fft_array.events)
        fft_events = []
        for j in range(self.iternumber):
                fft_events.append(self.fft.enqueue_arrays(
                    data=self._tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    result=self._tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    forward=True)[0])
        self._tmp_fft_array.add_event(
            self.prg.fftshift(
                self.queue,
                (self.fft_shape[0],
                 self.fft_shape[1],
                 self.fft_shape[2]),
                None,
                self._tmp_fft_array.data,
                self._check.data,
                wait_for=fft_events))
        # Resample on Spoke
        return self.prg.invgrid_lut(
            self.queue,
            (s.shape[0], s.shape[1] * s.shape[2], s.shape[-2] *
             self._gridsize),
            None,
            s.data,
            self._tmp_fft_array.data,
            self.traj.data,
            np.int32(self._gridsize),
            np.int32(s.shape[2]),
            self.DTYPE_real(self._kwidth),
            self.dcf.data,
            self.cl_kerneltable,
            np.int32(self._kernelpoints),
            np.int32(scan_offset),
            wait_for=s.events + wait_for + self._tmp_fft_array.events)


class PyOpenCL3DRadialNUFFT(PyOpenCLnuFFT):
    """Non-uniform FFT object.

    This class performs the 3D non-uniform FFT (NUFFT) operation. Linear
    interpolation of a sampled gridding kernel is used to regrid points
    from the non-cartesian grid back on the cartesian grid.

    Parameters
    ----------
      ctx : PyOpenCL.Context
        The context for the PyOpenCL computations.
      queue : PyOpenCL.Queue
        The computation Queue for the PyOpenCL kernels.
      par : dict
        A python dict containing the necessary information to
        setup the object. Needs to contain the number of slices (NSlice),
        number of scans (NScan), image dimensions (dimX, dimY), number of
        coils (NC), sampling points (N) and read outs (NProj)
        a PyOpenCL queue (queue) and the complex coil
        sensitivities (C).
      kwidth : int
        The width of the sampling kernel for regridding of non-uniform
        kspace samples.
      klength : int
        The length of the kernel lookup table which samples the contineous
        gridding kernel.
      DTYPE : Numpy.dtype
        The comlex precision type. Currently complex64 is used.
      DTYPE_real : Numpy.dtype
        The real precision type. Currently float32 is used.

    Attributes
    ----------
      traj : PyOpenCL.Array
        The comlex sampling trajectory
      dcf : PyOpenCL.Array
        The densitiy compenation function
      ogf (float):
        The overgriddingfactor for non-cartesian k-spaces.
      fft_shape : tuple of ints
        3 dimensional tuple. Dim 0 containts all Scans, Coils and Slices.
        Dim 1 and 2 the overgridded image dimensions.
      fft_scale : float32
        The scaling factor to achieve a good adjointness of the forward and
        backward FFT.
      cl_kerneltable (PyOpenCL.Buffer):
        The gridding lookup table as read only Buffer
      cl_deapo (PyOpenCL.Buffer):
        The deapodization lookup table as read only Buffer
      par_fft : int
        The number of parallel fft calls. Typically it iterates over the
        Scans.
      fft : gpyfft.fft.FFT
        The fft object created from gpyfft (A wrapper for clFFT). The object
        is created only once an reused in each iterations, iterationg over
        all scans to keep the memory footprint low.
      prg : PyOpenCL.Program
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator. This will be determined by the
        factory and set after the object is created.
    """

    def __init__(
            self,
            ctx,
            queue,
            par,
            kwidth=5,
            klength=200,
            DTYPE=np.complex64,
            DTYPE_real=np.float32,
            streamed=False):
        super().__init__(ctx, queue, par["fft_dim"], DTYPE, DTYPE_real)
        
        # self.ogf = par["N"]/par["dimX"]     
        self.ogf = par["ogf"]

        self.fft_shape = (
            par["NScan"] *
            par["NC"],
            int(round(par["NSlice"]*self.ogf)),
            int(round(par["dimY"]*self.ogf)),
            int(round(par["dimX"]*self.ogf)))

        self.fft_scale = DTYPE_real(
            np.sqrt(np.prod(self.fft_shape[-3:])))

        (kerneltable, kerneltable_FT) = calckbkernel(
            kwidth, self.ogf, par["N"], klength)

        deapo = 1 / kerneltable_FT.astype(DTYPE_real)

        self.cl_kerneltable = cl.Buffer(
            self.ctx,
            cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
            hostbuf=kerneltable.astype(DTYPE_real).data)
        self.cl_deapo = cl.Buffer(
            self.ctx,
            cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
            hostbuf=deapo.data)
        self.dcf = clarray.to_device(self.queue, par["dcf"])
        self.traj = clarray.to_device(self.queue, par["traj"])
        self._tmp_fft_array = (
            clarray.zeros(
                self.queue,
                (self.fft_shape),
                dtype=DTYPE))
        if par["use_GPU"]:
            self.par_fft = int(
                self.fft_shape[0] / par["NScan"])
        else:
            self.par_fft = self.fft_shape[0]
        self.iternumber = int(self.fft_shape[0]/self.par_fft)
        self.fft = FFT(ctx, queue, self._tmp_fft_array[
            0:self.par_fft, ...],
                       out_array=self._tmp_fft_array[
                           0:self.par_fft, ...],
                       axes=self.fft_dim)

        self._kernelpoints = kerneltable.size
        self._kwidth = kwidth / 2
        self._check = np.ones(self.fft_shape[-1], dtype=DTYPE_real)
        self._check[1::2] = -1
        self._check = clarray.to_device(self.queue, self._check)
        self._gridsize = self.fft_shape[-1]

    def __del__(self):
        """Explicitly delete OpenCL Objets."""
        del self.traj
        del self.dcf
        del self._tmp_fft_array
        del self.cl_kerneltable
        del self.cl_deapo
        del self._check
        del self.queue
        del self.ctx
        del self.prg
        del self.fft

    def FFTH(self, sg, s, wait_for=None, scan_offset=0):
        """Perform the inverse (adjoint) NUFFT operation.

        Parameters
        ----------
          sg : PyOpenCL.Array
            The complex image data.
          s : PyOpenCL.Array
            The non-uniformly gridded k-space
          wait_for : list of PyopenCL.Event, None
            A List of PyOpenCL events to wait for.
          scan_offset : int, 0
            Offset compared to the first acquired scan.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        if wait_for is None:
            wait_for = []
        # Zero tmp arrays
        self._tmp_fft_array.add_event(
            self.prg.zero_tmp(
                self.queue,
                (self._tmp_fft_array.size,
                 ),
                None,
                self._tmp_fft_array.data,
                wait_for=self._tmp_fft_array.events))

        # Grid k-space
        self._tmp_fft_array.add_event(
            self.prg.grid_lut3D(
                self.queue,
                (s.shape[0], s.shape[1],
                 s.shape[-2] * self._gridsize),
                None,
                self._tmp_fft_array.data,
                s.data,
                self.traj.data,
                np.int32(self._gridsize),
                np.int32(sg.shape[2]),
                self.DTYPE_real(self._kwidth),
                self.dcf.data,
                self.cl_kerneltable,
                np.int32(self._kernelpoints),
                np.int32(scan_offset),
                wait_for=(wait_for + 
                          s.events + self._tmp_fft_array.events)))
        # FFT
        self._tmp_fft_array.add_event(
            self.prg.fftshift3D(
                self.queue,
                (np.prod(self.fft_shape[:2]),
                 self.fft_shape[2],
                 self.fft_shape[3]),
                None,
                self._tmp_fft_array.data,
                self._check.data,
                wait_for=self._tmp_fft_array.events))  
        cl.wait_for_events(self._tmp_fft_array.events)
        fft_events = []
        for j in range(self.iternumber):
                fft_events.append(self.fft.enqueue_arrays(
                    data=self._tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    result=self._tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    forward=False)[0])
        self._tmp_fft_array.add_event(
            self.prg.fftshift3D(
                self.queue,
                (np.prod(self.fft_shape[:2]),
                 self.fft_shape[2],
                 self.fft_shape[3]),
                None,
                self._tmp_fft_array.data,
                self._check.data,
                wait_for=fft_events))
        return self.prg.deapo_adj3D(
            self.queue,
            (sg.shape[0] * sg.shape[1] *
             sg.shape[2], sg.shape[3], sg.shape[4]),
            None,
            sg.data,
            self._tmp_fft_array.data,
            self.cl_deapo,
            np.int32(self._tmp_fft_array.shape[-1]),
            self.DTYPE_real(self.fft_scale),
            self.DTYPE_real(self.ogf),
            wait_for=(wait_for + sg.events + 
                      self._tmp_fft_array.events))

    def FFT(self, s, sg, wait_for=None, scan_offset=0):
        """Perform the forward NUFFT operation.

        Parameters
        ----------
          s : PyOpenCL.Array
            The non-uniformly gridded k-space.
          sg : PyOpenCL.Array
            The complex image data.
          wait_for : list of PyopenCL.Event, None
            A List of PyOpenCL events to wait for.
          scan_offset : int, 0
            Offset compared to the first acquired scan.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        if wait_for is None:
            wait_for = []
        # Zero tmp arrays

        self._tmp_fft_array.add_event(
            self.prg.zero_tmp(
                self.queue,
                (self._tmp_fft_array.size,
                 ),
                None,
                self._tmp_fft_array.data,
                wait_for=
                self._tmp_fft_array.events))
        # Deapodization and Scaling
        self._tmp_fft_array.add_event(
            self.prg.deapo_fwd3D(
                self.queue,
                (sg.shape[0] * sg.shape[1] * sg.shape[2],
                 sg.shape[3], sg.shape[4]),
                None,
                self._tmp_fft_array.data,
                sg.data,
                self.cl_deapo,
                np.int32(self._tmp_fft_array.shape[-1]),
                self.DTYPE_real(1 / self.fft_scale),
                self.DTYPE_real(self.ogf),
                wait_for=wait_for + sg.events + self._tmp_fft_array.events))

        # FFT
        self._tmp_fft_array.add_event(
            self.prg.fftshift3D(
                self.queue,
                (np.prod(self.fft_shape[:2]),
                  self.fft_shape[2],
                  self.fft_shape[3]),
                None,
                self._tmp_fft_array.data,
                self._check.data,
                wait_for=self._tmp_fft_array.events))
        
        cl.wait_for_events(self._tmp_fft_array.events)
        fft_events = []
        for j in range(self.iternumber):
                fft_events.append(self.fft.enqueue_arrays(
                    data=self._tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    result=self._tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    forward=True)[0])

        self._tmp_fft_array.add_event(
            self.prg.fftshift3D(
                self.queue,
                (np.prod(self.fft_shape[:2]),
                 self.fft_shape[2],
                 self.fft_shape[3]),
                None,
                self._tmp_fft_array.data,
                self._check.data,
                wait_for=fft_events))

        # Resample on Spoke       
        return self.prg.invgrid_lut3D(
            self.queue,
            (s.shape[0], s.shape[1], s.shape[-2] *
             self._gridsize),
            None,
            s.data,
            self._tmp_fft_array.data,
            self.traj.data,
            np.int32(self._gridsize),
            np.int32(s.shape[2]),
            self.DTYPE_real(self._kwidth),
            self.dcf.data,
            self.cl_kerneltable,
            np.int32(self._kernelpoints),
            np.int32(scan_offset),
            wait_for=s.events + wait_for + self._tmp_fft_array.events)


class PyOpenCLCartNUFFT(PyOpenCLnuFFT):
    """Cartesian FFT object.

    This class performs the FFT operation.

    Parameters
    ----------
      ctx : PyOpenCL.Context
        The context for the PyOpenCL computations.
      queue : PyOpenCL.Queue
        The computation Queue for the PyOpenCL kernels.
      par : dict A python dict containing the necessary information to
        setup the object. Needs to contain the number of slices (NSlice),
        number of scans (NScan), image dimensions (dimX, dimY), number of
        coils (NC), sampling points (N) and read outs (NProj)
        a PyOpenCL queue (queue) and the complex coil
        sensitivities (C).
      DTYPE : Numpy.dtype
        The comlex precision type. Currently complex64 is used.
      DTYPE_real : Numpy.dtype
        The real precision type. Currently float32 is used.

    Attributes
    ----------
      fft_shape : tuple of ints
        3 dimensional tuple. Dim 0 containts all Scans, Coils and Slices.
        Dim 1 and 2 the overgridded image dimensions.
      fft_scale : float32
        The scaling factor to achieve a good adjointness of the forward and
        backward FFT.
      par_fft : int
        The number of parallel fft calls. Typically it iterates over the
        Scans.
      fft : gpyfft.fft.FFT
        The fft object created from gpyfft (A wrapper for clFFT). The object
        is created only once an reused in each iterations, iterationg over
        all scans to keep the memory footprint low.
      mask : PyOpenCL.Array
        The undersampling mask for the Cartesian grid.
      prg : PyOpenCL.Program
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator. This will be determined by the
        factory and set after the object is created.
    """

    def __init__(
            self,
            ctx,
            queue,
            par,
            DTYPE=np.complex64,
            DTYPE_real=np.float32,
            streamed=False):
        super().__init__(ctx, queue, par["fft_dim"], DTYPE, DTYPE_real)
        if streamed:
            self.fft_shape = (
                par["NScan"] *
                par["NC"] *
                (par["par_slices"] + par["overlap"]), 
                par["dimY"], 
                par["dimX"])
        else:
            if par["is3D"]:
                self.fft_shape = (
                    par["NScan"] *
                    par["NC"],
                    par["NSlice"],
                    par["dimY"],
                    par["dimX"])
            else:
                self.fft_shape = (
                    par["NScan"] *
                    par["NC"] *
                    par["NSlice"],
                    par["dimY"],
                    par["dimX"])

        if par["fft_dim"] is not None:
            self.fft_scale = DTYPE_real(
                np.sqrt(np.prod(self.fft_shape[self.fft_dim[0]:])))
            self._tmp_fft_array = (
                clarray.zeros(
                    self.queue,
                    self.fft_shape,
                    dtype=DTYPE))
            if par["use_GPU"]:
                self.par_fft = int(
                    self.fft_shape[0] / par["NScan"])
            else:
                self.par_fft = self.fft_shape[0]
            self.iternumber = int(self.fft_shape[0]/self.par_fft)
            self.mask = clarray.to_device(self.queue, par["mask"])
            self.fft = FFT(ctx, queue, self._tmp_fft_array[
                0:self.par_fft, ...],
                           out_array=self._tmp_fft_array[
                               0:self.par_fft, ...],
                           axes=self.fft_dim)

    def __del__(self):
        """Explicitly delete OpenCL Objets."""
        if self.fft_dim is not None:
            del self._tmp_fft_array
            del self.fft
            del self.mask
        del self.queue
        del self.ctx
        del self.prg

    def FFTH(self, sg, s, wait_for=None, scan_offset=0):
        """Perform the inverse (adjoint) FFT operation.

        Parameters
        ----------
          sg : PyOpenCL.Array
            The complex image data.
          s : PyOpenCL.Array
            The uniformly gridded k-space
          wait_for : list of PyopenCL.Event, None
            A List of PyOpenCL events to wait for.
          scan_offset : int, 0
            Offset compared to the first acquired scan.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        if wait_for is None:
            wait_for = []
        if self.fft_dim is not None:
            self._tmp_fft_array.add_event(
                self.prg.maskingcpy(
                    self.queue,
                    (self._tmp_fft_array.shape[0],
                     np.prod(self._tmp_fft_array.shape[1:])),
                    None,
                    self._tmp_fft_array.data,
                    s.data,
                    self.mask.data,
                    wait_for=s.events+self._tmp_fft_array.events+wait_for))
            cl.wait_for_events(self._tmp_fft_array.events)
            fft_events = []
            for j in range(self.iternumber):
                    fft_events.append(self.fft.enqueue_arrays(
                        data=self._tmp_fft_array[
                            j * self.par_fft:(j + 1) * self.par_fft, ...],
                        result=self._tmp_fft_array[
                            j * self.par_fft:(j + 1) * self.par_fft, ...],
                        forward=False)[0])
            return (
                self.prg.copy(
                    self.queue,
                    (sg.size,
                     ),
                    None,
                    sg.data,
                    self._tmp_fft_array.data,
                    self.DTYPE_real(
                        self.fft_scale),
                    wait_for=sg.events+fft_events))

        return self.prg.copy(
                    self.queue,
                    (sg.size,
                     ),
                    None,
                    sg.data,
                    s.data,
                    self.DTYPE_real(1),
                    wait_for=s.events+sg.events+wait_for)

    def FFT(self, s, sg, wait_for=None, scan_offset=0):
        """Perform the forward FFT operation.

        Parameters
        ----------
          s : PyOpenCL.Array
            The uniformly gridded k-space.
          sg : PyOpenCL.Array
            The complex image data.
          wait_for : list of PyopenCL.Event, None
            A List of PyOpenCL events to wait for.
          scan_offset : int, 0
            Offset compared to the first acquired scan.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        if wait_for is None:
            wait_for = []
        if self.fft_dim is not None:
            self._tmp_fft_array.add_event(
                self.prg.copy(
                    self.queue,
                    (sg.size,
                     ),
                    None,
                    self._tmp_fft_array.data,
                    sg.data,
                    self.DTYPE_real(
                        1 /
                        self.fft_scale),
                    wait_for=sg.events+self._tmp_fft_array.events+wait_for))
            cl.wait_for_events(self._tmp_fft_array.events)
            fft_events = []
            for j in range(self.iternumber):
                    fft_events.append(self.fft.enqueue_arrays(
                        data=self._tmp_fft_array[
                            j * self.par_fft:(j + 1) * self.par_fft, ...],
                        result=self._tmp_fft_array[
                            j * self.par_fft:(j + 1) * self.par_fft, ...],
                        forward=True)[0])

            return (
                self.prg.maskingcpy(
                    self.queue,
                    (self._tmp_fft_array.shape[0],
                     np.prod(self._tmp_fft_array.shape[1:])),
                    None,
                    s.data,
                    self._tmp_fft_array.data,
                    self.mask.data,
                    wait_for=s.events+fft_events))

        return self.prg.copy(
                    self.queue,
                    (sg.size,
                     ),
                    None,
                    s.data,
                    sg.data,
                    self.DTYPE_real(1),
                    wait_for=s.events+sg.events+wait_for)


class PyOpenCLSMSNUFFT(PyOpenCLnuFFT):
    """Cartesian FFT-SMS object.

    This class performs the FFT operation assuming a SMS acquisition.

    Parameters
    ----------
      ctx : PyOpenCL.Context
        The context for the PyOpenCL computations.
      queue : PyOpenCL.Queue
        The computation Queue for the PyOpenCL kernels.
      par : dict A python dict containing the necessary information to
        setup the object. Needs to contain the number of slices (NSlice),
        number of scans (NScan), image dimensions (dimX, dimY), number of
        coils (NC), sampling points (N) and read outs (NProj)
        a PyOpenCL queue (queue) and the complex coil
        sensitivities (C).
      DTYPE : Numpy.dtype
        The comlex precision type. Currently complex64 is used.
      DTYPE_real : Numpy.dtype
        The real precision type. Currently float32 is used.

    Attributes
    ----------
      fft_shape : tuple of ints
        3 dimensional tuple. Dim 0 containts all Scans, Coils and Slices.
        Dim 1 and 2 the overgridded image dimensions.
      fft_scale : float32
        The scaling factor to achieve a good adjointness of the forward and
        backward FFT.
      par_fft : int
        The number of parallel fft calls. Typically it iterates over the
        Scans.
      fft : gpyfft.fft.FFT
        The fft object created from gpyfft (A wrapper for clFFT). The object
        is created only once an reused in each iterations, iterationg over
        all scans to keep the memory footprint low.
      mask : PyOpenCL.Array
        The undersampling mask for the Cartesian grid.
      packs : int
        The distance between the slices
      MB : int
        The multiband factor
      shift : PyOpenCL.Array
        The vector pixel shifts used in the fft computation.
      prg : PyOpenCL.Program
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator. This will be determined by the
        factory and set after the object is created.
    """

    def __init__(
            self,
            ctx,
            queue,
            par,
            DTYPE=np.complex64,
            DTYPE_real=np.float32,
            streamed=False):
        super().__init__(ctx, queue, par["fft_dim"], DTYPE, DTYPE_real)
        
        if streamed:
            self.fft_shape = (
                par["NC"] *
                par["NSlice"],
                par["dimY"],
                par["dimX"])
        else:
            self.fft_shape = (
                par["NScan"] *
                par["NC"] *
                par["NSlice"],
                par["dimY"],
                par["dimX"])

        self.packs = int(par["packs"])
        self.MB = int(par["MB"])
        self.shift = clarray.to_device(
            self.queue, par["shift"].astype(DTYPE_real))

        if par["fft_dim"] is not None:
            self.fft_scale = DTYPE_real(
                np.sqrt(np.prod(self.fft_shape[self.fft_dim[0]:])))
            self._tmp_fft_array = (
                clarray.zeros(
                    self.queue,
                    self.fft_shape,
                    dtype=DTYPE))
            if par["use_GPU"] and not streamed:
                self.par_fft = int(
                    self.fft_shape[0] / par["NScan"])
            else:
                self.par_fft = self.fft_shape[0]
            self.iternumber = int(self.fft_shape[0]/self.par_fft)
            self.mask = clarray.to_device(self.queue, par["mask"])
            self.fft = FFT(ctx, queue, self._tmp_fft_array[
                0:self.par_fft, ...],
                           out_array=self._tmp_fft_array[
                               0:self.par_fft, ...],
                           axes=self.fft_dim)

    def __del__(self):
        """Explicitly delete OpenCL Objets."""
        if self.fft_dim is not None:
            del self._tmp_fft_array
            del self.fft
        del self.mask
        del self.queue
        del self.ctx
        del self.prg

    def FFTH(self, sg, s, wait_for=None, scan_offset=0):
        """Perform the inverse (adjoint) FFT operation.

        Parameters
        ----------
          sg : PyOpenCL.Array
            The complex image data.
          s : PyOpenCL.Array
            The uniformly gridded k-space compressed by the MB factor.
          wait_for : list of PyopenCL.Event, None
            A List of PyOpenCL events to wait for.
          scan_offset : int, 0
            Offset compared to the first acquired scan.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        if wait_for is None:
            wait_for = []
        if self.fft_dim is not None:
            self._tmp_fft_array.add_event(
                self.prg.copy_SMS_adjkspace(
                    self.queue,
                    (sg.shape[0] * sg.shape[1],
                     sg.shape[-2],
                     sg.shape[-1]),
                    None,
                    self._tmp_fft_array.data,
                    s.data,
                    self.shift.data,
                    self.mask.data,
                    np.int32(self.packs),
                    np.int32(self.MB),
                    self.DTYPE_real(self.fft_scale),
                    np.int32(sg.shape[2]/self.packs/self.MB),
                    wait_for=s.events+wait_for+self._tmp_fft_array.events))
            cl.wait_for_events(self._tmp_fft_array.events)
            fft_events = []
            for j in range(self.iternumber):
                    fft_events.append(self.fft.enqueue_arrays(
                        data=self._tmp_fft_array[
                            j * self.par_fft:(j + 1) * self.par_fft, ...],
                        result=self._tmp_fft_array[
                            j * self.par_fft:(j + 1) * self.par_fft, ...],
                        forward=False)[0])

            return (self.prg.copy(self.queue,
                                  (sg.size,),
                                  None,
                                  sg.data,
                                  self._tmp_fft_array.data,
                                  self.DTYPE_real(self.fft_scale),
                                  wait_for=sg.events+fft_events))

        return self.prg.copy_SMS_adj(
                self.queue,
                (sg.shape[0] * sg.shape[1],
                 sg.shape[-2],
                 sg.shape[-1]),
                None,
                sg.data,
                s.data,
                self.shift.data,
                self.mask.data,
                np.int32(self.packs),
                np.int32(self.MB),
                self.DTYPE_real(1),
                np.int32(sg.shape[2]/self.packs/self.MB),
                wait_for=s.events+sg.events+wait_for)

    def FFT(self, s, sg, wait_for=None, scan_offset=0):
        """Perform the forward FFT operation.

        Parameters
        ----------
          s : PyOpenCL.Array
            The uniformly gridded k-space compressed by the MB factor.
          sg : PyOpenCL.Array
            The complex image data.
          wait_for : list of PyopenCL.Event, None
            A List of PyOpenCL events to wait for.
          scan_offset : int, 0
            Offset compared to the first acquired scan.

        Returns
        -------
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        if wait_for is None:
            wait_for = []
        if self.fft_dim is not None:
            self._tmp_fft_array.add_event(
                self.prg.copy(
                    self.queue,
                    (sg.size,),
                    None,
                    self._tmp_fft_array.data,
                    sg.data,
                    self.DTYPE_real(1 / self.fft_scale),
                    wait_for=self._tmp_fft_array.events+sg.events+wait_for))
            cl.wait_for_events(self._tmp_fft_array.events)
            fft_events = []
            for j in range(self.iternumber):
                    fft_events.append(self.fft.enqueue_arrays(
                        data=self._tmp_fft_array[
                            j * self.par_fft:(j + 1) * self.par_fft, ...],
                        result=self._tmp_fft_array[
                            j * self.par_fft:(j + 1) * self.par_fft, ...],
                        forward=True)[0])

            return (
                self.prg.copy_SMS_fwdkspace(
                    self.queue,
                    (s.shape[0] * s.shape[1], s.shape[-2], s.shape[-1]),
                    None,
                    s.data,
                    self._tmp_fft_array.data,
                    self.shift.data,
                    self.mask.data,
                    np.int32(self.packs),
                    np.int32(self.MB),
                    self.DTYPE_real(self.fft_scale),
                    np.int32(sg.shape[2]/self.packs/self.MB),
                    wait_for=s.events+fft_events+wait_for))

        return (
            self.prg.copy_SMS_fwd(
                self.queue,
                (s.shape[0] * s.shape[1], s.shape[-2], s.shape[-1]),
                None,
                s.data,
                sg.data,
                self.shift.data,
                self.mask.data,
                np.int32(self.packs),
                np.int32(self.MB),
                self.DTYPE_real(1),
                np.int32(sg.shape[2]/self.packs/self.MB),
                wait_for=s.events+sg.events+wait_for))
